import os 
import json 
import pandas as pd
import warnings 

def read_json(filename, **kwargs):
    """
    read in a JSON format file.  return None if the file is not there.

    Parameters
    ----------
    filename: string
        name of the JSON file to read
    except: bool
        if true, raise exception if file is missing. if false, return empty dict.

    Returns
    -------
    d: python object
        data from JSON file decoded into python object
    """
    try:
        with open(filename, 'r') as f:
            json_data = json.load(f, **kwargs)
    except:
        msg = "Missing JSON file: %s" % filename
        raise Exception(msg)
    d = json_data
    return d


def read_hdf(filename, requires, raise_except=False, **kwargs):
    """
    read in H5 format file and pull out specific table based on specific
    header information in the requires field. return None if file is not there 

    Parameters
    ----------
    filename: str
        name of H5 file to read 
    requires : dict 
        dictionary with keys and values to get correct table 
        e.g. {'wmin':0.3, 'wmax':1, 'R':5000} for opacity grid

    Returns
    -------
    d : python object 
        data from H5 file decoded into python object 
    """
    requires_list = [i+'='+str(requires[i]) for i in requires.keys()]

    hdf_name = pd.read_hdf(filename,'header',where=requires_list, columns=['table'],**kwargs)

    if len(hdf_name['table'].values) > 1:
        msg = "HDF requirements satify two grids. Using %s in %s" % (hdf_name.values[0],filename)
        warnings.warn(msg, UserWarning)
        d = pd.read_hdf(filename, hdf_name['table'].values[0])
    else: 
        d = pd.read_hdf(filename, hdf_name['table'].values[0])
    return d



from .rayleigh import Rayleigh

import pandas as pd
import numpy as np
import json
import os
from numba import jit
from bokeh.plotting import figure, show, output_file
from bokeh.palettes import inferno
import io 
import sqlite3
import math

#@jit(nopython=True)
def compute_opacity(atmosphere, opacityclass, stream, delta_eddington=True,test_mode=False,raman=0, plot_opacity=False,
    full_output=False):
    """
    Returns total optical depth per slab layer including molecular opacity, continuum opacity. 
    It should automatically select the molecules needed
    
    Parameters
    ----------
    atmosphere : class ATMSETUP
        This inherets the class from atmsetup.py 
    opacityclass : class opacity
        This inherets the class from optics.py. It is done this way so that the opacity db doesnt have 
        to be reloaded in a retrieval 
    delta_eddington : bool 
        (Optional) Default=True, With Delta-eddington on, it incorporates the forward peak 
        contribution by adjusting optical properties such that the fraction of scattered energy
        in the forward direction is removed from the scattering parameters 
    raman : int 
        (Optional) Default =0 which corresponds to oklopcic+2018 raman scattering cross sections. 
        Other options include 1 for original pollack approximation for a 6000K blackbody. 
        And 2 for nothing.  
    test_mode : bool 
        (Optional) Default = False to run as normal. This will overwrite the opacities and fix the 
        delta tau at 0.5. 
    full_output : bool 
        (Optional) Default = False. If true, This will add taugas, taucld, tauray to the atmosphere class. 
        This is done so that the users can debug, plot, etc. 
    plot_opacity : bool 
        (Optional) Default = False. If true, Will create a pop up plot of the weighted of each absorber 
        at the middle layer.

    Returns
    -------
    DTAU : ndarray 
        This is a matrix with # layer by # wavelength. It is the opacity contained within a layer 
        including the continuum, scattering, cloud (if specified), and molecular opacity
        **If requested, this is corrected for with Delta-Eddington.**
    TAU : ndarray
        This is a matrix with # level by # wavelength. It is the cumsum of opacity contained 
        including the continuum, scattering, cloud (if specified), and molecular opacity
        **If requested, this is corrected for with Delta-Eddington.**
    WBAR : ndarray
        This is the single scattering albedo that includes rayleigh, raman and user input scattering sources. 
        It has dimensions: # layer by # wavelength
        **If requested, this is corrected for with Delta-Eddington.**
    COSB : ndarray
        This is the asymettry factor which accounts for rayleigh and user specified values 
        It has dimensions: # layer by # wavelength
        **If requested, this is corrected for with Delta-Eddington.**
    ftau_cld : ndarray 
        This is the fraction of cloud opacity relative to the total TAUCLD/(TAUCLD + TAURAY)
    ftau_ray : ndarray 
        This is the fraction of rayleigh opacity relative to the total TAURAY/(TAUCLD + TAURAY)
    GCOS2 : ndarray
        This is used for Cahoy+2010 methodology for accounting for rayleigh scattering. It 
        replaces the use of the actual rayleigh phase function by just multiplying ftau_ray by 2
    DTAU : ndarray 
        This is a matrix with # layer by # wavelength. It is the opacity contained within a layer 
        including the continuum, scattering, cloud (if specified), and molecular opacity
        **If requested, this is corrected for with Delta-Eddington.**
    TAU : ndarray
        This is a matrix with # level by # wavelength. It is the cumsum of opacity contained 
        including the continuum, scattering, cloud (if specified), and molecular opacity
        **Original, never corrected for with Delta-Eddington.**
    WBAR : ndarray
        This is the single scattering albedo that includes rayleigh, raman and user input scattering sources. 
        It has dimensions: # layer by # wavelength
        **Original, never corrected for with Delta-Eddington.**
    COSB : ndarray
        This is the asymettry factor which accounts for rayleigh and user specified values 
        It has dimensions: # layer by # wavelength
        **Original, never corrected for with Delta-Eddington.**

    Notes
    -----
    This was baselined against jupiter with the old fortran code. It matches 100% for all cases 
    except for hotter cases where Na & K are present. This differences is not a product of the code 
    but a product of the different opacities (1060 grid versus old 736 grid)

    Todo 
    -----
    Add a better approximation than delta-scale (e.g. M.Marley suggests a paper by Cuzzi that has 
    a better methodology)
    """
    atm = atmosphere
    tlevel = atm.level['temperature']
    #these units are purely because of the wonky continuum units
    #plevel in this routine is only used for those 
    plevel = atm.level['pressure']/atm.c.pconv #THIS IS DANGEROUS, used for continuum
    
    tlayer = atm.layer['temperature']
    player = atm.layer['pressure']/atm.c.pconv #THIS IS DANGEROUS, used for continuum
    gravity = atm.planet.gravity / 100.0  #THIS IS DANGEROUS, used for continuum
    nlayer = atm.c.nlayer
    nwno = opacityclass.nwno

    if plot_opacity: 
        plot_layer=int(nlayer/1.5)#np.size(tlayer)-1
        opt_figure = figure(x_axis_label = 'Wavelength', y_axis_label='TAUGAS in optics.py', 
        title = 'Opacity at T='+str(tlayer[plot_layer])+' P='+str(player[plot_layer])
        ,y_axis_type='log',height=700, width=600)

    #====================== INITIALIZE TAUGAS#======================
    TAUGAS = 0 
    TAURAY = 0 
    c=1
    #set color scheme.. adding 3 for raman, rayleigh, and total
    if plot_opacity: colors = inferno(3+len(atm.continuum_molecules) + len(atm.molecules))

    #====================== ADD CONTIMUUM OPACITY====================== 
    #Set up coefficients needed to convert amagat to a normal human unit
    #these COEF's are only used for the continuum opacity. 
    ACOEF = (tlayer/(tlevel[:-1]*tlevel[1:]))*(
            tlevel[1:]*plevel[1:] - tlevel[:-1]*plevel[:-1])/(plevel[1:]-plevel[:-1]) #UNITLESS

    BCOEF = (tlayer/(tlevel[:-1]*tlevel[1:]))*(
            tlevel[:-1] - tlevel[1:])/(plevel[1:]-plevel[:-1]) #INVERSE PRESSURE

    COEF1 = atm.c.rgas*273.15**2*.5E5* (
        ACOEF* (plevel[1:]**2 - plevel[:-1]**2) + BCOEF*(
            2./3.)*(plevel[1:]**3 - plevel[:-1]**3) ) / (
        1.01325**2 *gravity*tlayer*atm.layer['mmw']) 

    #go through every molecule in the continuum first 
    for m in atm.continuum_molecules:

        #H- Bound-Free
        if (m[0] == "H-") and (m[1] == "bf"):
            ADDTAU = (opacityclass.continuum_opa['H-bf']*(           #[(nwno x nlayer) *(
                            atm.layer['mixingratios'][m[0]].values*  #nlayer
                            atm.layer['colden']/                     #nlayer
                            (atm.layer['mmw']*atm.c.amu))   ).T      #nlayer)].T

            TAUGAS += ADDTAU
            if plot_opacity: opt_figure.line(1e4/opacityclass.wno, ADDTAU[plot_layer,:], alpha=0.7,legend_label=m[0]+m[1], line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2)

        #H- Free-Free
        elif (m[0] == "H-") and (m[1] == "ff"):
            ADDTAU = (opacityclass.continuum_opa['H-ff']*(                               #[(nwno x nlayer) *(
                            atm.layer['pressure']*                                       #nlayer
                            atm.layer['mixingratios']['H'].values*atm.layer['electrons']*#nlayer
                            atm.layer['colden']/                                         #nlayer
                            (tlayer*atm.layer['mmw']*atm.c.amu*atm.c.k_b))  ).T          #nlayer)].T
            #testing['H-ff'] = ADDTAU
            TAUGAS += ADDTAU
            if plot_opacity: opt_figure.line(1e4/opacityclass.wno, ADDTAU[plot_layer,:], alpha=0.7,legend_label=m[0]+m[1], line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2)

        #H2- 
        elif (m[0] == "H2-") and (m[1] == ""): 
            #calculate opacity
            #this is a hefty matrix multiplication to make sure that we are 
            #multiplying each column of the opacities by the same 1D vector (as opposed to traditional 
            #matrix multiplication). This is the reason for the transposes.
            ADDTAU = (opacityclass.continuum_opa['H2-']*(               #[(nwno x nlayer) *(
                            atm.layer['pressure']*                                      #nlayer
                            atm.layer['mixingratios']['H2'].values*atm.layer['electrons']*  #nlayer
                            atm.layer['colden']/                                        #nlayer
                            (atm.layer['mmw']*atm.c.amu))   ).T                         #nlayer)].T


            TAUGAS += ADDTAU
            if plot_opacity: opt_figure.line(1e4/opacityclass.wno, ADDTAU[plot_layer,:], alpha=0.7,legend_label=m[0]+m[1], line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2)
        #everything else.. e.g. H2-H2, H2-CH4. Automatically determined by which molecules were requested
        else:

            #calculate opacity
            ADDTAU = (opacityclass.continuum_opa[m[0]+m[1]] * ( #[(nwno x nlayer) *(
                                COEF1*                                          #nlayer
                                atm.layer['mixingratios'][m[0]].values *                #nlayer
                                atm.layer['mixingratios'][m[1]].values )  ).T           #nlayer)].T

            TAUGAS += ADDTAU
            if plot_opacity: opt_figure.line(1e4/opacityclass.wno, ADDTAU[plot_layer,:], alpha=0.7,legend_label=m[0]+m[1], line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2)
        c+=1
    
    #====================== ADD MOLECULAR OPACITY====================== 
    for m in atm.molecules:
        #if m =='O3':
        #    df = pd.read_csv('~/Desktop/LIFE/O3_visible.txt',delim_whitespace=True,names=['nm','cx'])
        #    wno_old1 = (df['nm']*1e-3).values
        #    wno1 = 1e4/opacityclass.wno[::-1]
        #    o31 = np.zeros((len(wno1),1))
        #    o31[:,0] = np.interp(wno1, wno_old1 ,df['cx'], left=1e-33, right=1e-33)[::-1]*6.02214086e+23 
        #    opacityclass.molecular_opa[m] += o31

        ADDTAU = (opacityclass.molecular_opa[m] * ( #[(nwno x nlayer) *(
                    atm.layer['colden']*
                    atm.layer['mixingratios'][m].values/ #removing this bc of opa unit change *atm.weights[m].values[0]/ 
                    atm.layer['mmw']) ).T 
        TAUGAS += ADDTAU
        #testing[m] = ADDTAU
        if plot_opacity: opt_figure.line(1e4/opacityclass.wno, ADDTAU[plot_layer,:], alpha=0.7,legend_label=m, line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2)
        c+=1

    #====================== ADD RAYLEIGH OPACITY======================  
    #old 1
    """
    ray_mixingratios = np.zeros((nlayer,3))#hardwired because we only have h2,he and ch4 scattering
    for i,j in zip(['H2','He','CH4'],range(3)):
        if i in atm.rayleigh_molecules:
            print(i)
            ray_mixingratios[:,j] = atm.layer['mixingratios'][i].values
    
    TAURAY = rayleigh_old(atm.layer['colden'],ray_mixingratios, 
                    opacityclass.wave, atm.layer['mmw'],atm.c.amu )
    """
    for m in atm.rayleigh_molecules:
        ray_matrix = np.array([opacityclass.rayleigh_opa[m]]*nlayer).T
        ADDTAU = (ray_matrix * ( #[(nwno x nlayer) *(
                    atm.layer['colden']*
                    atm.layer['mixingratios'][m].values/ #removing this bc of opa unit change *atm.weights[m].values[0]/ 
                    atm.layer['mmw'])).T 
        TAURAY += ADDTAU 
        
    """
    #old 3 
    lam = np.zeros((len(opacityclass.wno),1))
    lam[:,0] = 1e4/opacityclass.wno
    nam1=1E-8*(8342.13+2406030./(130.-lam**(-2))+15997./(38.9-lam**(-2)))
    ff=1.05
    Na0=2.55E19*100.**3
    ray_matrix=np.zeros((len(opacityclass.wno), nlayer))+32.*np.pi**3./(3.*Na0**2)*(nam1**2)*ff/((lam*1E-6)**4.)
    TAURAY = (6.02214086e+23 * 1e4 * ray_matrix * ( #[(nwno x nlayer) *(
                    atm.layer['colden']*
                    atm.layer['mixingratios']['N2'].values/ #removing this bc of opa unit change *atm.weights[m].values[0]/ 
                    atm.layer['mmw'])).T 
    """
    if plot_opacity: opt_figure.line(1e4/opacityclass.wno, TAURAY[plot_layer,:], alpha=0.7,legend_label='Rayleigh', line_width=3, color=colors[c],
            muted_color=colors[c], muted_alpha=0.2) 


    #====================== ADD RAMAN OPACITY======================
    #OKLOPCIC OPACITY
    if raman == 0 :
        raman_db = opacityclass.raman_db
        raman_factor = compute_raman(nwno, nlayer,opacityclass.wno, 
            opacityclass.raman_stellar_shifts, tlayer, raman_db['c'].values,
                raman_db['ji'].values, raman_db['deltanu'].values)
        if plot_opacity: opt_figure.line(1e4/opacityclass.wno, raman_factor[plot_layer,:]*TAURAY[plot_layer,:], alpha=0.7,legend_label='Shifted Raman', line_width=3, color=colors[c],
                muted_color=colors[c], muted_alpha=0.2)
        raman_factor = np.minimum(raman_factor, raman_factor*0+0.99999)
    #POLLACK OPACITY
    elif raman ==1: 
        raman_factor = raman_pollack(nlayer,1e4/opacityclass.wno)
        raman_factor = np.minimum(raman_factor, raman_factor*0+0.99999)  
        if plot_opacity: opt_figure.line(1e4/opacityclass.wno, raman_factor[plot_layer,:]*TAURAY[plot_layer,:], alpha=0.7,legend_label='Shifted Raman', line_width=3, color=colors[c],
                muted_color=colors[c], muted_alpha=0.2)
    #NOTHING
    else: 
        raman_factor = 0.99999
    

    #====================== ADD CLOUD OPACITY====================== 
    TAUCLD = atm.layer['cloud']['opd'] #TAUCLD is the total extinction from cloud = (abs + scattering)
    asym_factor_cld = atm.layer['cloud']['g0']
    single_scattering_cld = atm.layer['cloud']['w0'] 

    #====================== If user requests full output, add Tau's to atmosphere class=====
    if full_output:
        atmosphere.taugas = TAUGAS
        atmosphere.tauray = TAURAY
        atmosphere.taucld = TAUCLD

    #====================== ADD EVERYTHING TOGETHER PER LAYER====================== 
    #formerly DTAU
    DTAU = TAUGAS + TAURAY + TAUCLD 
    
    # This is the fractional of the total scattering that will be due to the cloud
    #VERY important note. You must weight the taucld by the single scattering 
    #this is because we only care about the fractional opacity from the cloud that is 
    #scattering. 
    ftau_cld = (single_scattering_cld * TAUCLD)/(single_scattering_cld * TAUCLD + TAURAY)

    COSB = ftau_cld*asym_factor_cld

    #formerly GCOSB2 
    ftau_ray = TAURAY/(TAURAY + single_scattering_cld * TAUCLD)
    GCOS2 = 0.5*ftau_ray #Hansen & Travis 1974 for Rayleigh scattering 

    #Raman correction is usually used for reflected light calculations 
    #although users to have option turn it off in reflected light as well 
    W0 = (TAURAY*raman_factor + TAUCLD*single_scattering_cld) / (TAUGAS + TAURAY + TAUCLD) #TOTAL single scattering 

    #if a user wants both reflected and thermal, this computes SSA without raman correction, but with 
    #scattering from clouds still
    W0_no_raman = (TAURAY*0.99999 + TAUCLD*single_scattering_cld) / (TAUGAS + TAURAY + TAUCLD) #TOTAL single scattering 

    #sum up taus starting at the top, going to depth
    shape = DTAU.shape
    TAU = np.zeros((shape[0]+1, shape[1]))
    TAU[1:,:]=numba_cumsum(DTAU)

    if plot_opacity:
        opt_figure.line(1e4/opacityclass.wno, DTAU[plot_layer,:], legend_label='TOTAL', line_width=4, color=colors[0],
            muted_color=colors[c], muted_alpha=0.2)
        opt_figure.legend.click_policy="mute"
        show(opt_figure)

    if test_mode != None:  
            #this is to check against Dlugach & Yanovitskij 
            #https://www.sciencedirect.com/science/article/pii/0019103574901675?via%3Dihub
            if test_mode=='rayleigh':
                DTAU = TAURAY 
                #GCOS2 = 0.5
                GCOS2 = np.zeros(DTAU.shape) + 0.5
                #ftau_ray = 1.0
                ftau_ray = np.zeros(DTAU.shape) + 1.0
                #ftau_cld = 1e-6
                ftau_cld = np.zeros(DTAU.shape) #+ 1e-6
            else: 
                DTAU = atm.layer['cloud']['opd']#TAURAY*0+0.05
                GCOS2 = np.zeros(DTAU.shape)#0.0
                ftau_ray = np.zeros(DTAU.shape)#+1e-6
                ftau_cld = np.zeros(DTAU.shape)+1            
            COSB = atm.layer['cloud']['g0']
            W0 = atm.layer['cloud']['w0']
            W0_no_raman = atm.layer['cloud']['w0']
            TAU = np.zeros((shape[0]+1, shape[1]))
            TAU[1:,:]=numba_cumsum(DTAU)

    #====================== D-Eddington Approximation======================
    if delta_eddington:
        #First thing to do is to use the delta function to icorporate the forward 
        #peak contribution of scattering by adjusting optical properties such that 
        #the fraction of scattered energy in the forward direction is removed from 
        #the scattering parameters 

        #Joseph, J.H., W. J. Wiscombe, and J. A. Weinman, 
        #The Delta-Eddington approximation for radiative flux transfer, J. Atmos. Sci. 33, 2452-2459, 1976.

        #also see these lecture notes are pretty good
        #http://irina.eas.gatech.edu/EAS8803_SPRING2012/Lec20.pdf
        w0_dedd=W0*(1.-COSB**stream)/(1.0-W0*COSB**stream)
        cosb_dedd=COSB/(1.+COSB)
        dtau_dedd=DTAU*(1.-W0*COSB**stream) 

        #sum up taus starting at the top, going to depth
        tau_dedd = np.zeros((shape[0]+1, shape[1]))
        tau_dedd[1:,:]=numba_cumsum(dtau_dedd)
    
        #returning the terms used in 
        return (dtau_dedd, tau_dedd, w0_dedd, cosb_dedd ,ftau_cld, ftau_ray, GCOS2, 
                DTAU, TAU, W0, COSB,    #these are returned twice because we need the uncorrected 
                W0_no_raman)            #values for single scattering terms where we use the TTHG phase function
                                        # w0_no_raman is used in thermal calculations only

    else: 
        return (DTAU, TAU, W0, COSB, ftau_cld, ftau_ray, GCOS2, 
                DTAU, TAU, W0, COSB,  #these are returned twice for consistency with the delta-eddington option
                W0_no_raman)          #W0_no_raman is used for thermal calculations only 

@jit(nopython=True, cache=True)
def rayleigh_old(colden,gasmixing,wave,xmu,amu):
    """DISCONTINUED
    Rayleigh function taken from old albedo code. Keeping this modular, as we may want 
    to swap out different methods to calculate rayleigh opacity 

    Parameters
    ----------
    colden : array of float 
        Column Density in CGS units 
    mixingratios : array of float 
        Mixing ratio as function of altitude for [H2  He CH4 ] 
    wave : array of float 
        wavelength (microns) of grid. This should be in DECENDING order to comply with 
        everything else being in wave numberes 
    xmu : array of float
        mean molecular weight of atmosphere in amu 
    amu : float 
        amu constant in grams 
    """
    #define all rayleigh constants
    TAURAY = np.zeros((colden.size, wave.size))
    dpol= np.zeros(3)
    dpol[0], dpol[1], dpol[2]=1.022 , 1.0, 1.0
    gnu = np.zeros((2,3))
    gnu[0,0]=1.355e-4
    gnu[1,0]=1.235e-6
    gnu[0,1]=3.469e-5
    gnu[1,1]=8.139e-8
    gnu[0,2]=4.318e-4
    gnu[1,2]=3.408e-6
    XN0 = 2.687E19
    cfray = 32.0*np.pi**3*1.e21/(3.0*2.687e19)
    cold = colden / (xmu * amu) #nlayers
    #add rayleigh from each contributing gas using corresponding mixing 
    for i in np.arange(0,3,1):
        tec = cfray*(dpol[i]/wave**4)*(gnu[0,i]+gnu[1,i]/   #nwave
                     wave**2)**2 
        TAUR = (cold*gasmixing[:,i]).reshape((-1, 1)) * tec * 1e-5 / XN0
        
        TAURAY += TAUR

    return TAURAY

@jit(nopython=True, cache=True)
def compute_raman(nwno, nlayer, wno, stellar_shifts, tlayer, cross_sections, j_initial, deltanu):
    """
    The Ramam scattering will alter the rayleigh scattering. The returned value is 
    modified single scattering albedo. 

    Cross sectiosn from: 
    http://iopscience.iop.org/0004-637X/832/1/30/suppdata/apjaa3ec7t2_mrt.txt

    This method is described in Pollack+1986. Albeit not the best method. Sromovsky+2005 
    pointed out the inconsistencies in this method. You can see from his comparisons 
    that the Pollack approximations don't accurately capture the depths of the line centers. 
    Since then, OKLOPCIC+2016 recomputed cross sections for J<=9. We are using those cross 
    sections here with a different star. Huge improvement over what we had before. 
    Number of J levels is hard coded to 10 ! 

    Will be added to the rayleigh scattering as : TAURAY*RAMAN

    Parameters
    ----------
    nwno : int 
        Number of wave points
    nlayer : int 
        Number of layers 
    wno : array
        Array of output grid of wavenumbers
    stellar_shifts : ndarray 
        Array of floats that has dimensions (n wave pts x n J levels: 0-9)
    tlayer : array 
        Array of floats that has dimensions nlayer 
    cross_sections : ndarray 
        The row of "C's" from Antonija's table. 
    j_initial : ndarray 
        The row of initial rotational energy states from Antonija's table
    deltanu : ndarray
        The row of delta nu's from Antonia's table

    """
    raman_sigma_w_shift = np.zeros(( nlayer,nwno))
    raman_sigma_wo_shift = np.zeros(( nlayer,nwno))
    rayleigh_sigma = np.zeros(( nlayer,nwno))

    #first calculate the j fraction at every layer 

    number_of_Js = 10 #(including 0)
    j_at_temp = np.zeros((number_of_Js,nlayer))
    for i in range(number_of_Js):
        j_at_temp[i,:] = j_fraction(i,tlayer)
    
    for i in range(cross_sections.shape[0]):
        #define initial state
        ji = j_initial[i]
        #for that initial state.. What is the expected wavenumber shift 
        shifted_wno = wno + deltanu[i]
        #compute the cross section 
        Q = cross_sections[i] / wno**3.0 / shifted_wno #see A4 in Antonija's 2018 paper
        #if deltanu is zero that is technically rayleigh scattering
        if deltanu[i] == 0 :
            rayleigh_sigma += np.outer(j_at_temp[ji,:] , Q )
        else:
            #if not, then compute the shifted and unshifted raman scattering
            raman_sigma_w_shift += np.outer(j_at_temp[ji,:] , Q*stellar_shifts[:,i])
            raman_sigma_wo_shift += np.outer(j_at_temp[ji,:] , Q)

    #finally return the contribution that will be added to total rayleigh
    return (rayleigh_sigma + raman_sigma_w_shift)/ (rayleigh_sigma + raman_sigma_wo_shift)

@jit(nopython=True, cache=True)
def bin_star(wno_new,wno_old, Fp):
    """
    Takes average of group of points using uniform tophat 
    
    Parameters
    ----------
    wno_new : numpy.array
        inverse cm grid to bin to (cm-1) 
    wno_old : numpy.array
        inverse cm grid (old grid)
    Fp : numpy.array
        transmission spectra, which is on wno grid
    """
    szmod=wno_new.shape[0]

    delta=np.zeros(szmod)
    Fint=np.zeros(szmod)
    delta[0:-1]=wno_new[1:]-wno_new[:-1]  
    delta[szmod-1]=delta[szmod-2] 
    for i in range(1,szmod):
        loc=np.where((wno_old >= wno_new[i]-0.5*delta[i-1]) & (wno_old < wno_new[i]+0.5*delta[i]))
        Fint[i]=np.mean(Fp[loc])
    loc=np.where((wno_old > wno_new[0]-0.5*delta[0]) & (wno_old < wno_new[0]+0.5*delta[0]))
    Fint[0]=np.mean(Fp[loc])
    return Fint


@jit(nopython=True, cache=True)
def partition_function(j, T):
    """
    Given j and T computes partition function for ro-vibrational levels of H2. 
    This is the exponential and the statistical weight g_J in 
    Eqn 3 in https://arxiv.org/pdf/1605.07185.pdf
    It is also used to compute the partition sum Z.

    Parameters
    ----------
    j : int 
        Energy level 
    T : array float 
        Temperature at each atmospheric layer in the atmosphere

    Returns
    -------
    Returns partition function 
    """
    k = 1.38064852e-16  #(c.k_B.cgs)
    b = 60.853# units of /u.cm #rotational constant for H2
    c =  29979245800 #.c.cgs
    h = 6.62607004e-27 #c.h.cgs
    b_energy = (b*(h)*(c)*j*(j+1)/k)
    if (j % 2 == 0 ):
        return (2.0*j+1.0)*np.exp(-0.5*b_energy*j*(j+1)/T) #1/2 is the symmetry # for homonuclear molecule
    else:
        return 3.0*(2.0*j+1.0)*np.exp(-0.5*b_energy*j*(j+1)/T)

@jit(nopython=True, cache=True)
def partition_sum(T):
    """
    This is the total partition sum. I am truncating it at 20 since it seems to approach 1 around then. 
    This is also pretty fast to compute so 20 is fine for now. 

    Parameters
    ----------
    T : array 
        Array of temperatures for each layer 

    Returns
    -------
    Z, the partition sum 
    """
    Z=np.zeros(T.size)
    for j in range(0,20):
        Z += partition_function(j,T)
    return Z

@jit(nopython=True, cache=True)
def j_fraction(j,T):
    """
    This computes the fraction of molecules at each J level. 

    Parameters
    ----------
    j : int 
        The initial rotational levels ranging from J=0 to 9 for hydrogen only

    Returns
    -------
    f_J in eqn. 3 https://arxiv.org/pdf/1605.07185.pdf
    """
    return partition_function(j,T)/partition_sum(T)

#@jit(nopython=True, cache=True)
def raman_pollack(nlayer,wave):
    """
    Mystery raman scattering. Couldn't figure out where it came from.. so discontinuing. 
    Currently function doesnt' totally work. In half fortran-half python. Legacy from 
    fortran albedo code. 

    This method is described in Pollack+1986. Albeit not the best method. Sromovsky+2005 
    pointed out the inconsistencies in this method. You can see from his comparisons 
    that the Pollack approximations don't accurately capture the depths of the line centers. 
    Since then, OKLOPCIC+2016 did a much
    better model of raman scattring (with ghost lines). Might be worth it to consider a more 
    sophisticated version of Raman scattering. 

    Will be added to the rayleigh scattering as : TAURAY*RAMAN
    
    #OLD FORTRAN CODE
    #constants 
    h = 6.6252e-27
    c = 2.9978e10
    bohrd = 5.2917e-9
    hmass = 1.6734e-24
    rmu = .5 * hmass

    #set wavelength shift of the ramam scatterer
    shift_v0 = 4161.0 

    facip = h * c / ( 1.e-4 * 27.2 * 1.602e-12 ) 
    facray = 1.e16 * bohrd ** 3 * 128. * np.pi ** 5 * bohrd ** 3 / 9. 
    facv = 2.08 / 2.38 * facray / bohrd ** 2 / ( 8. * np.pi * np.pi * rmu * c * shift_v0 ) * h

    #cross section of the unshifted rayleigh and the vibrationally shifted rayleigh
    gli = np.zeros(5)
    wli = np.zeros(5) 
    gri = np.zeros(5) 
    wri = np.zeros(5) 
    gli[:] = [1.296, .247, .297,  .157,  .003]
    wli[:] = [.507, .628, .733, 1.175, 2.526]
    gri[:] = [.913, .239, .440,  .344,  .064]
    wri[:] = [.537, .639, .789, 1.304, 3.263]

    alp = np.zeros(7)
    arp = np.zeros(7)
    alp[:] = [6.84, 6.96, 7.33, 8.02, 9.18, 11.1, 14.5 ]
    arp[:] = [3.66, 3.71, 3.88, 4.19, 4.70, 5.52, 6.88 ]

    omega = facip / wavelength

    #first compute extinction cross section for unshifted component 
    #e.g. rayleigh
    alpha_l=0
    alpha_r=0

    for i in range(5):
        alpha_l += gli[i] / ( wli[i] ** 2 - omega ** 2 ) 
        alpha_r += gri[i] / ( wri[i] ** 2 - omega ** 2 )

    alpha2 = (( 2. * alpha_r + alpha_l ) / 3. ) ** 2
    gamma2 = ( alpha_l - alpha_r ) ** 2
    qray = facray * ( 3. * alpha2 + 2./3. * gamma2 ) / wavelength ** 4

    #next, compute the extinction cross section for vibrationally 
    #shifted component 
    ip = np.zeros(2)
    ip[:] = [int(omega/0.05), 5.0]
    ip = np.min(ip) + 1 
    f = omega / 0.05 - float(ip-1)
    alpha_pl = ( 1. - f ) * alp[ip] + f * alp[ip+1]
    alpha_pr = ( 1. - f ) * arp[ip] + f * arp[ip+1]
    alpha_p2 = (( 2. * alpha_pr + alpha_pl ) / 3. ) ** 2
    gamma_p2 = ( alpha_pl - alpha_pr ) ** 2
    qv = facv / SHIFT( WAVEL, -SHIFTV0 ) ** 4 * ( 3. * alpha_p2 + 2./3. * gamma_p2 )    
    """
    dat = pd.read_csv(os.path.join(os.environ.get('picaso_refdata'), 'opacities','raman_fortran.txt'),
                        delim_whitespace=True, header=None, names = ['w','f'])
    #fill in matrix to match real raman format
    interp_raman = np.interp(wave, dat['w'].values, dat['f'].values, )
    raman_factor = np.zeros((nlayer, len(wave)))
    for i in range(nlayer): 
        raman_factor[i,:] = interp_raman#return flipped values for raman
    return  raman_factor 

class RetrieveOpacities():
    """
    This will be the class that will retrieve the opacities from the sqlite3 database. Right 
    now we are just employing nearest neighbors to grab the respective opacities. 
    Eventually, we will switch to correlated K. 
    
    Parameters
    ----------
    db_filename : str 
        Opacity file name 
    raman_data : str 
        file name of the raman cross section data 
    wave_range : list of float
        Wavelength range (in microns) to run. 
        Default None grabs all the full wavelength range available 
    location : str 
        Default = local. This is the only option currently available since 
        we dont have AWS or other services enabled. 
    resample : int 
        Default =1 (no resampling)

    Attributes
    ----------
    raman_db : pandas.DataFrame
        Database of Raman cross sections 
    wno : array 
        Wavenumber grid from opacity database (CGS cm-1)
    wave : array 
        Wavelength grid from opacity database (micron)
    nwno : int 
        Number of wavenumber points in the grid 
    cia_temps : array 
        CIA temps available from database 
    molecules : array 
        Molecules availalbe from database 
    pt_pairs : array 
        pressure-temperature pairs available from grid (our grid is usually 
        computed on a fixed 1060 PT grid)
    molecular_opa : dict 
        Molecular Opacity dictionary that contains all the queried information necessary 
        to do calculation 
    continuum_opa : dict 
        Continuum opacity dictionary that contains all the queried information 
        necessary for the calculation 
    rayleigh_opa : dict 
        Rayleigh opacity dictionary for all the sources we have. This is computed as 
        cm^2/species when the users runs `opannection`. We technically could just 
        store this in the DB as well. However, because its so fast, it doesn't take 
        a lot of time, and because its non-temperature dependent, we only have to do 
        it once. 

    Methods 
    -------
    db_connect 
        Inherents functionality of `open_local`. Eventually it will 
        inherent functionality of "open_non_locals" (e.g. cloud or other sources)
    open_local 
        Opens sqlite3 connection and enables array interpretation from bytes 
    get_available_data 
        Gets opacity database attributes such as available T-Ps , availalbe molecules. etc. 
    get_available_rayleigh
        Gets rayleigh cross sections when opannection is opened.
    get_opacities 
        Gets opacities after user specifies atmospheric profile (e.g. full PT and Composition)
    """
    #def __init__(self, continuum_data, molecular_data,raman_data, db = 'local'):
    def __init__(self, db_filename, raman_data, wave_range=None,location = 'local',resample=1):

        if location == 'local':
            #self.conn = sqlite3.connect(db_filename, detect_types=sqlite3.PARSE_DECLTYPES)
            self.db_filename = db_filename
            self.db_connect = self.open_local

        self.get_available_data(wave_range, resample)
        
        #raman cross sections 
        self.raman_db = pd.read_csv(raman_data,
                     delim_whitespace=True, skiprows=16,header=None, names=['ji','jf','vf','c','deltanu'])
        
        #compute available Rayleigh scatterers 
        self.get_available_rayleigh()

    def open_local(self):
        """Code needed to open up local database, interpret arrays from bytes and return cursor"""
        conn = sqlite3.connect(self.db_filename, detect_types=sqlite3.PARSE_DECLTYPES)
        #tell sqlite what to do with an array
        sqlite3.register_adapter(np.ndarray, self.adapt_array)
        sqlite3.register_converter("array", self.convert_array)
        cur = conn.cursor()
        return cur,conn

    def get_available_data(self, wave_range, resample):
        """Get the pressures and temperatures that are available for the continuum and molecules"""
        self.resample = resample
        #open connection 
        cur, conn = self.db_connect()

        #get temps
        cur.execute('SELECT temperature FROM continuum')
        self.cia_temps = np.unique(cur.fetchall())

        #get available molecules
        cur.execute('SELECT molecule FROM molecular')
        self.molecules = np.unique(cur.fetchall())

        #get PT indexes for getting opacities 
        cur.execute('SELECT ptid, pressure, temperature FROM molecular')
        data= cur.fetchall()
        self.pt_pairs = sorted(list(set(data)),key=lambda x: (x[0]) )

        #Get the wave grid info
        cur.execute('SELECT wavenumber_grid FROM header')
        self.wno =  cur.fetchone()[0][::self.resample]
        self.wave = 1e4/self.wno 
        if wave_range == None: 
            self.loc = [True]*len(self.wno )
        else:
            self.loc = np.where(((self.wave>min(wave_range)) & (self.wave<max(wave_range))))
        self.wave = self.wave[self.loc]
        self.wno = self.wno[self.loc]
        self.nwno = np.size(self.wno)

        conn.close()

    def get_available_rayleigh(self):
        data = Rayleigh(self.wno)
        self.rayleigh_molecules = data.rayleigh_molecules
        self.rayleigh_opa={}
        for i in data.rayleigh_molecules: 
            self.rayleigh_opa[i] = data.compute_sigma(i)

    def get_opacities(self,atmosphere, dimension='1d'):
        """
        Get's opacities using the atmosphere class
        """
        #open connection 
        cur, conn = self.db_connect()
        
        #if dimension=='1d':
        nlayer =atmosphere.c.nlayer
        tlayer =atmosphere.layer['temperature']
        player = atmosphere.layer['pressure']/atmosphere.c.pconv
        #elif dimension=='3d':
        #   nlayer =atmosphere.c.nlayer
        #   tlayer =atmosphere.layer['temperature'].ravel()
        #   player = (atmosphere.layer['pressure']/atmosphere.c.pconv).ravel()

        molecules = atmosphere.molecules
        cia_molecules = atmosphere.continuum_molecules

        #struture opacity dictionary
        self.molecular_opa = {key:np.zeros((self.nwno, nlayer)) for key in molecules}
        self.continuum_opa = {key[0]+key[1]:np.zeros((self.nwno, nlayer)) for key in cia_molecules}

        #this will make getting opacities faster 
        #this is getting the ptid corresponding to the pairs

        ind_pt=[min(self.pt_pairs, 
            key=lambda c: math.hypot(c[1]- coordinate[0], c[2]-coordinate[1]))[0] 
                for coordinate in  zip(player,tlayer)]
        #query molecular opacities from sqlite3
        if len(molecules) ==1: 
            query_mol = """WHERE molecule= '{}' """.format(str(molecules[0]))
        else:
            query_mol = 'WHERE molecule in '+str(tuple(molecules) )

        atmosphere.layer['pt_opa_index'] = ind_pt
        cur.execute("""SELECT molecule,ptid,opacity 
                    FROM molecular 
                    {} 
                    AND ptid in {}""".format(query_mol, str(tuple(np.unique(ind_pt)))))
        #fetch everything and stick into a dictionary where we can find the right
        #pt and molecules
        data= cur.fetchall()
        data = dict((x+'_'+str(y), dat[::self.resample][self.loc]) for x,y,dat in data)        

        #structure it into a dictionary e.g. {'H2O':ndarray(nwave x nlayer), 'CH4':ndarray(nwave x nlayer)}.. 
        for i in self.molecular_opa.keys():
            for j,ind in zip(ind_pt,range(nlayer)):
                self.molecular_opa[i][:,ind] = data[i+'_'+str(j)]*6.02214086e+23 #add to opacity bundle

        #continuum
        #find nearest temp for cia grid
        tcia = [np.unique(self.cia_temps)[find_nearest(np.unique(self.cia_temps),i)] for i in tlayer]

        #if user only runs a single molecule or temperature
        if len(tcia) ==1: 
            query_temp = """AND temperature= '{}' """.format(str(tcia[0]))
        else:
            query_temp = 'AND temperature in '+str(tuple(tcia) )
        cia_mol = list(self.continuum_opa.keys())
        if len(cia_mol) ==1: 
            query_mol = """WHERE molecule= '{}' """.format(str(cia_mol[0]))
        else:
            query_mol = 'WHERE molecule in '+str(tuple(cia_mol) )       

        cur.execute("""SELECT molecule,temperature,opacity 
                    FROM continuum 
                    {} 
                    {}""".format(query_mol, query_temp))

        data = cur.fetchall()
        data = dict((x+'_'+str(y), dat) for x, y,dat in data)
        for i in self.continuum_opa.keys():
            for j,ind in zip(tcia,range(nlayer)):
                self.continuum_opa[i][:,ind] = data[i+'_'+str(j)][::self.resample][self.loc]

        conn.close() 
             
    def get_continuum_opac(self, temperature, molecule): 
        """DISCONTINUED.
        Based on a temperature, this retrieves the continuum opacity for 

        Parameters
        ----------
        temperature : float array
            An array of temperatures to retrieve the continuum opacity at

        molecule : str 
            Which opacity source to query. currently available: 'h2h2','h2he','h2h','h2ch4','h2n2'

        Return
        ------
        matrix 
            number of layers by number of wave points 
        """
        nearestT = [self.cia_temps[find_nearest(self.cia_temps, t)] for t in temperature]
        sizeT = np.size(nearestT)
        a = np.zeros((sizeT, self.nwno))
        for i,t in zip(range(sizeT) ,nearestT): 
            self.cur.execute('SELECT opacity FROM continuum WHERE molecule=? AND temperature=?',(molecule,t))
            a[i,:] = self.cur.fetchall()[0][0]#np.array(self.cia_db[molecule][t])
        return a

    def get_molecular_opac(self, temperature, pressure, molecule):
        """DISCONTINUED.
        Based on temperature, and pressure, retrieves the molecular opacity for 
        certain molecule 

        Molecular opacity should be in cm2/molecule units. That is why there is an 
        multiplication of N_a on return 

        Parameters
        ----------
        temperature : array of float 
            An array of temperatures to retrieve the continuum opacity at 
        pressure : array of float 
            Pressures (IN BARS) to retrieve the continuum opacity at 
        molecule : str 
            Which opacity source to query. Will get error if not in the db
        """
        #if (molecule == 'Na') or (molecule == 'K'):
        #   mol_temps = np.array([float(i) for i in self.mol_db['Na'].keys()])
        #   mol_press = {}
        #   for i in mol_temps:
        #       mol_press[str(i)] = np.array([float(i) for i in self.mol_db['Na'][str(i)].keys()])
        #   nearestT = [mol_temps[find_nearest(mol_temps, t)].astype(str) for t in temperature]
        #   nearestP = [mol_press[str(t)][find_nearest(mol_press[str(t)], p)].astype(str) for p,t in zip(pressure,nearestT)]
        #else:
        nearestT = [self.mol_temps[find_nearest(self.mol_temps, t)] for t in temperature]
        sizeT = np.size(nearestT)
        a = np.zeros((sizeT, self.nwno))
        for i,t,p in zip(range(sizeT) ,nearestT,pressure): 
            self.cur.execute('SELECT opacity FROM molecular WHERE molecule=? AND temperature=? ORDER BY ABS(? - pressure) LIMIT 1',(molecule,t,p))
            a[i,:] = self.cur.fetchall()[0][0]
        return a*6.02214086e+23 

    def compute_stellar_shits(self, wno_star, flux_star):
        """
        Takes the hires stellar spectrum and computes Raman shifts 
        
        Parameters
        ----------
        wno_star : array 
            hires wavelength grid of the star in wave number
        flux_star : array 
            Flux of the star in whatever units. Doesn't matter for this since it's returning a ratio
        
        Returns
        -------
        matrix
            array of the shifted stellar spec divided by the unshifted stellar spec on the model wave
            number grid 
        """
        model_wno = self.wno
        deltanu = self.raman_db['deltanu'].values

        all_shifted_spec = np.zeros((len(model_wno), len(deltanu)))
        
        self.unshifted_stellar_spec = bin_star(model_wno, wno_star, flux_star)

        for i in range(len(deltanu)):
            dnu = deltanu[i]
            shifted_wno = model_wno + dnu 
            shifted_flux = bin_star(shifted_wno,wno_star, flux_star)
            if i ==0:
                unshifted = shifted_flux*0+shifted_flux
            all_shifted_spec[:,i] = shifted_flux/unshifted

        self.raman_stellar_shifts = all_shifted_spec

    def adapt_array(arr):
        """needed to interpret bytes to array"""
        out = io.BytesIO()
        np.save(out, arr)
        out.seek(0)
        return sqlite3.Binary(out.read())

    def convert_array(clas, text):
        """needed to interpret bytes to array"""
        out = io.BytesIO(text)
        out.seek(0)
        return np.load(out)

@jit(nopython=True, cache=True)
def find_nearest(array,value):
    #small program to find the nearest neighbor in temperature  
    idx = (np.abs(array-value)).argmin()
    return idx
@jit(nopython=True, cache=True)
def numba_cumsum(mat):
    """Function to compute cumsum along axis=0 to bypass numba not allowing kwargs in 
    cumsum 
    """
    new_mat = np.zeros(mat.shape)
    for i in range(mat.shape[1]):
        new_mat[:,i] = np.cumsum(mat[:,i])
    return new_mat



from .atmsetup import ATMSETUP
from .fluxes import get_reflected_1d, get_reflected_3d , get_thermal_1d, get_thermal_3d, get_reflected_new, get_transit_1d
from .wavelength import get_cld_input_grid
from .opacity_factory import create_grid
from .optics import RetrieveOpacities,compute_opacity
from .disco import get_angles_1d, get_angles_3d, compute_disco, compress_disco, compress_thermal

from virga import justdoit as vj
from scipy.signal import savgol_filter
from scipy.interpolate import RegularGridInterpolator
import scipy as sp
from scipy import special
from scipy.stats import binned_statistic

import requests
import os
import pickle as pk
import numpy as np
import pandas as pd
import copy
import json
import pysynphot as psyn
import astropy.units as u
import astropy.constants as c
import math

__refdata__ = os.environ.get('picaso_refdata')

def picaso(bundle,opacityclass, dimension = '1d',calculation='reflected', full_output=False, 
    plot_opacity= False, as_dict=True):
    """
    Currently top level program to run albedo code 

    Parameters 
    ----------
    bundle : dict 
        This input dict is built by loading the input = `justdoit.load_inputs()` 
    opacityclass : class
        Opacity class from `justdoit.opannection`
    dimension : str 
        (Optional) Dimensions of the calculation. Default = '1d'. But '3d' is also accepted. 
        In order to run '3d' calculations, user must build 3d input (see tutorials)
    full_output : bool 
        (Optional) Default = False. Returns atmosphere class, which enables several 
        plotting capabilities. 
    plot_opacity : bool 
        (Optional) Default = False, Creates pop up of the weighted opacity
    as_dict : bool 
        (Optional) Default = True. If true, returns a condensed dictionary to the user. 
        If false, returns the atmosphere class, which can be used for debugging. 
        The class is clunky to navigate so if you are consiering navigating through this, ping one of the 
        developers. 

    Return
    ------
    dictionary with albedos or fluxes or both (depending on what calculation type)
    """
    inputs = bundle.inputs

    wno = opacityclass.wno
    nwno = opacityclass.nwno

    #check to see if we are running in test mode
    test_mode = inputs['test_mode']

    ############# DEFINE ALL APPROXIMATIONS USED IN CALCULATION #############
    #see class `inputs` attribute `approx`

    #set approx numbers options (to be used in numba compiled functions)
    single_phase = inputs['approx']['single_phase']
    multi_phase = inputs['approx']['multi_phase']
    raman_approx =inputs['approx']['raman']
    method = inputs['approx']['method']
    stream = inputs['approx']['stream']
    tridiagonal = 0 

    #parameters needed for the two term hg phase function. 
    #Defaults are set in config.json
    f = inputs['approx']['TTHG_params']['fraction']
    frac_a = f[0]
    frac_b = f[1]
    frac_c = f[2]
    constant_back = inputs['approx']['TTHG_params']['constant_back']
    constant_forward = inputs['approx']['TTHG_params']['constant_forward']

    #define delta eddington approximinations 
    delta_eddington = inputs['approx']['delta_eddington']

    #pressure assumption
    p_reference =  inputs['approx']['p_reference']

    ############# DEFINE ALL GEOMETRY USED IN CALCULATION #############
    #see class `inputs` attribute `phase_angle`
    

    #phase angle 
    phase_angle = inputs['phase_angle']
    #get geometry
    geom = inputs['disco']

    ng, nt = geom['num_gangle'], geom['num_tangle']
    gangle,gweight,tangle,tweight = geom['gangle'], geom['gweight'],geom['tangle'], geom['tweight']
    lat, lon = geom['latitude'], geom['longitude']
    cos_theta = geom['cos_theta']
    ubar0, ubar1 = geom['ubar0'], geom['ubar1']

    #set star parameters
    radius_star = inputs['star']['radius']
    F0PI = np.zeros(nwno) + 1.
    #semi major axis
    sa = inputs['star']['semi_major']

    #begin atm setup
    atm = ATMSETUP(inputs)

    #Add inputs to class 
    atm.surf_reflect = inputs['surface_reflect']
    atm.wavenumber = wno
    atm.planet.gravity = inputs['planet']['gravity']
    atm.planet.radius = inputs['planet']['radius']
    atm.planet.mass = inputs['planet']['mass']

    if dimension == '1d':
        atm.get_profile()
    elif dimension == '3d':
        atm.get_profile_3d()

    #now can get these 
    atm.get_mmw()
    atm.get_density()
    atm.get_altitude(p_reference = p_reference)#will calculate altitude if r and m are given (opposed to just g)
    atm.get_column_density()

    #gets both continuum and needed rayleigh cross sections 
    #relies on continuum molecules are added into the opacity 
    #database. Rayleigh molecules are all in `rayleigh.py` 
    atm.get_needed_continuum(opacityclass.rayleigh_molecules)

    #get cloud properties, if there are any and put it on current grid 
    atm.get_clouds(wno)

    #Make sure that all molecules are in opacityclass. If not, remove them and add warning
    no_opacities = [i for i in atm.molecules if i not in opacityclass.molecules]
    atm.add_warnings('No computed opacities for: '+','.join(no_opacities))
    atm.molecules = np.array([ x for x in atm.molecules if x not in no_opacities ])

    nlevel = atm.c.nlevel
    nlayer = atm.c.nlayer
    

    if dimension == '1d':
        #lastly grab needed opacities for the problem
        opacityclass.get_opacities(atm)
        #only need to get opacities for one pt profile

        #There are two sets of dtau,tau,w0,g in the event that the user chooses to use delta-eddington
        #We use HG function for single scattering which gets the forward scattering/back scattering peaks 
        #well. We only really want to use delta-edd for multi scattering legendre polynomials. 
        DTAU, TAU, W0, COSB,ftau_cld, ftau_ray,GCOS2, DTAU_OG, TAU_OG, W0_OG, COSB_OG, W0_no_raman= compute_opacity(
            atm, opacityclass, stream, delta_eddington=delta_eddington,test_mode=test_mode,raman=raman_approx,
            full_output=full_output, plot_opacity=plot_opacity)

        if  'reflected' in calculation:
            #use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 
            nlevel = atm.c.nlevel
            if method == 'SH':
                xint_at_top = get_reflected_new(nlevel, nwno, ng, nt, 
                                            DTAU, TAU, W0, COSB, GCOS2, ftau_cld, ftau_ray,
                                            DTAU_OG, TAU_OG, W0_OG, COSB_OG, 
                                            atm.surf_reflect, ubar0, ubar1, cos_theta, F0PI, 
                                            single_phase, multi_phase, 
	                                    frac_a, frac_b, frac_c, constant_back, constant_forward, 
                                            dimension, stream, print_time)

            else:
                xint_at_top = get_reflected_1d(nlevel, wno,nwno,ng,nt,
                                                    DTAU, TAU, W0, COSB,GCOS2,ftau_cld,ftau_ray,
                                                    DTAU_OG, TAU_OG, W0_OG, COSB_OG ,
                                                    atm.surf_reflect, ubar0,ubar1,cos_theta, F0PI,
                                                    single_phase,multi_phase,
                                                    frac_a,frac_b,frac_c,constant_back,constant_forward, tridiagonal)

            #if full output is requested add in xint at top for 3d plots
            if full_output: 
                atm.xint_at_top = xint_at_top


        if 'thermal' in calculation:
            #use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 

            #remember all OG values (e.g. no delta eddington correction) go into thermal as well as 
            #the uncorrected raman single scattering 
            flux_at_top  = get_thermal_1d(nlevel, wno,nwno,ng,nt,atm.level['temperature'],
                                                    DTAU_OG, W0_no_raman, COSB_OG, atm.level['pressure'],ubar1,
                                                    atm.surf_reflect, tridiagonal)

            #if full output is requested add in flux at top for 3d plots
            if full_output: 
                atm.flux_at_top = flux_at_top
        
        if 'transmission' in calculation:
            rprs2 = get_transit_1d(atm.level['z'],atm.level['dz'],
                                  nlevel, nwno, radius_star, atm.layer['mmw'], 
                                  atm.c.k_b, atm.c.amu, atm.level['pressure'], 
                                  atm.level['temperature'], atm.layer['colden'],DTAU_OG)
    elif dimension == '3d':

        #setup zero array to fill with opacities
        TAU_3d = np.zeros((nlevel, nwno, ng, nt))
        DTAU_3d = np.zeros((nlayer, nwno, ng, nt))
        W0_3d = np.zeros((nlayer, nwno, ng, nt))
        COSB_3d = np.zeros((nlayer, nwno, ng, nt))
        GCOS2_3d = np.zeros((nlayer, nwno, ng, nt))
        FTAU_CLD_3d = np.zeros((nlayer, nwno, ng, nt))
        FTAU_RAY_3d = np.zeros((nlayer, nwno, ng, nt))

        #these are the unchanged values from delta-eddington
        TAU_OG_3d = np.zeros((nlevel, nwno, ng, nt))
        DTAU_OG_3d = np.zeros((nlayer, nwno, ng, nt))
        W0_OG_3d = np.zeros((nlayer, nwno, ng, nt))
        COSB_OG_3d = np.zeros((nlayer, nwno, ng, nt))
        #this is the single scattering without the raman correction 
        #used for the thermal caclulation
        W0_no_raman_3d = np.zeros((nlayer, nwno, ng, nt))

        #pressure and temperature 
        TLEVEL_3d = np.zeros((nlevel, ng, nt))
        PLEVEL_3d = np.zeros((nlevel, ng, nt))

        #if users want to retain all the individual opacity info they can here 
        if full_output:
            TAUGAS_3d = np.zeros((nlayer, nwno, ng, nt))
            TAUCLD_3d = np.zeros((nlayer, nwno, ng, nt))
            TAURAY_3d = np.zeros((nlayer, nwno, ng, nt))  

        #get opacities at each facet
        for g in range(ng):
            for t in range(nt): 

                #edit atm class to only have subsection of 3d stuff 
                atm_1d = copy.deepcopy(atm)

                #diesct just a subsection to get the opacity 
                atm_1d.disect(g,t)

                opacityclass.get_opacities(atm_1d)

                dtau, tau, w0, cosb,ftau_cld, ftau_ray, gcos2, DTAU_OG, TAU_OG, W0_OG, COSB_OG, WO_no_raman = compute_opacity(
                    atm_1d, opacityclass,delta_eddington=delta_eddington,test_mode=test_mode,raman=raman_approx, full_output=full_output)
                DTAU_3d[:,:,g,t] = dtau
                TAU_3d[:,:,g,t] = tau
                W0_3d[:,:,g,t] = w0 
                COSB_3d[:,:,g,t] = cosb
                GCOS2_3d[:,:,g,t]= gcos2 
                FTAU_CLD_3d[:,:,g,t]= ftau_cld
                FTAU_RAY_3d[:,:,g,t]= ftau_ray

                #these are the unchanged values from delta-eddington
                TAU_OG_3d[:,:,g,t] = TAU_OG
                DTAU_OG_3d[:,:,g,t] = DTAU_OG
                W0_OG_3d[:,:,g,t] = W0_OG
                COSB_OG_3d[:,:,g,t] = COSB_OG
                W0_no_raman_3d[:,:,g,t] = WO_no_raman

                #temp and pressure on 3d grid
                TLEVEL_3d[:,g,t] = atm_1d.level['temperature']
                PLEVEL_3d[:,g,t] = atm_1d.level['pressure']

                if full_output:
                    TAUGAS_3d[:,:,g,t] = atm_1d.taugas
                    TAUCLD_3d[:,:,g,t] = atm_1d.taucld
                    TAURAY_3d[:,:,g,t] = atm_1d.tauray

        if full_output:
            atm.taugas = TAUGAS_3d
            atm.taucld = TAUCLD_3d
            atm.tauray = TAURAY_3d

        if  'reflected' in calculation:
            #use toon method (and tridiagonal matrix solver) to get net cumulative fluxes 
            xint_at_top  = get_reflected_3d(nlevel, wno,nwno,ng,nt,
                                            DTAU_3d, TAU_3d, W0_3d, COSB_3d,GCOS2_3d, FTAU_CLD_3d,FTAU_RAY_3d,
                                            DTAU_OG_3d, TAU_OG_3d, W0_OG_3d, COSB_OG_3d,
                                            atm.surf_reflect, ubar0,ubar1,cos_theta, F0PI,
                                            single_phase,multi_phase,
                                            frac_a,frac_b,frac_c,constant_back,constant_forward, tridiagonal)
            #if full output is requested add in xint at top for 3d plots
            if full_output: 
                atm.xint_at_top = xint_at_top

        elif 'thermal' in calculation:

            #remember all OG values (e.g. no delta eddington correction) go into thermal as well as 
            #the uncorrected raman single scattering 
            flux_at_top  = get_thermal_3d(nlevel, wno,nwno,ng,nt,TLEVEL_3d,
                                                    DTAU_OG_3d, W0_no_raman_3d, COSB_OG_3d, PLEVEL_3d,ubar1, tridiagonal)

            #if full output is requested add in flux at top for 3d plots
            if full_output: 
                atm.flux_at_top = flux_at_top


    #COMPRESS FULL TANGLE-GANGLE FLUX OUTPUT ONTO 1D FLUX GRID

    #set up initial returns
    returns = {}
    returns['wavenumber'] = wno
    #returns['flux'] = flux
    if 'transmission' in calculation: 
        returns['transit_depth'] = rprs2


    #for reflected light use compress_disco routine
    #this takes the intensity as a functin of tangle/gangle and creates a 1d spectrum
    if  ('reflected' in calculation):
        albedo = compress_disco(nwno, cos_theta, xint_at_top, gweight, tweight,F0PI)
        returns['albedo'] = albedo 
        if ((not np.isnan(sa ) and (not np.isnan(atm.planet.radius))) ):
            returns['fpfs_reflected'] = albedo*(atm.planet.radius/sa)**2.0
        else: 
            returns['fpfs_reflected'] =[]
            if np.isnan(sa ):
                returns['fpfs_reflected'] += ['Semi-major axis not supplied. If you want fpfs, add it to `star` function. ']
            if np.isnan(atm.planet.radius): 
                returns['fpfs_reflected'] += ['Planet Radius not supplied. If you want fpfs, add it to `gravity` function with Mass.']

    #for thermal light use the compress thermal routine
    #this takes the intensity as a functin of tangle/gangle and creates a 1d spectrum
    if ('thermal' in calculation):
        thermal = compress_thermal(nwno,ubar1, flux_at_top, gweight, tweight)
        returns['thermal'] = thermal
        if full_output: atm.thermal_flux_planet = thermal

        #only need to return relative flux if not a browndwarf calculation
        if ((radius_star != 'nostar') & (not np.isnan(atm.planet.radius))):
            fpfs_thermal = thermal/(opacityclass.unshifted_stellar_spec)*(atm.planet.radius/radius_star)**2.0
            returns['fpfs_thermal'] = fpfs_thermal
        elif np.isnan(atm.planet.radius): 
            returns['fpfs_thermal'] = ['Planet Radius not supplied. If you want fpfs, add it to `gravity` function with Mass.']

    #return total if users have calculated both thermal and reflected 
    if (('fpfs_reflected' in list(returns.keys())) & ('fpfs_thermal' in list(returns.keys()))): 
        if ((not isinstance(returns['fpfs_reflected'],list)) & (not isinstance(returns['fpfs_thermal'],list))) :
            returns['fpfs_total'] = returns['fpfs_thermal'] + returns['fpfs_reflected']

    if full_output: 
        if as_dict:
            returns['full_output'] = atm.as_dict()
        else:
            returns['full_output'] = atm

    return returns

def opannection(wave_range = None, filename_db = None, raman_db = None, resample=1):
    """
    Sets up database connection to opacities. 

    Parameters
    ----------
    wave_range : list of float 
        Subset of wavelength range for which to run models for 
        Default : None, which pulls entire grid 
    filename_db : str 
        Filename of opacity database to query from 
        Default is none which pulls opacity file that comes with distribution 
    raman_db : str 
        Filename of raman opacity cross section 
        Default is none which pulls opacity file that comes with distribution 
    resample : int 
        Default=1 (no resampling) PROCEED WITH CAUTION!!!!!This will resample your opacites. 
        This effectively takes opacity[::BINS] depending on what the 
        sampling requested is. Consult your local theorist before 
        using this. 
    """

    inputs = json.load(open(os.path.join(__refdata__,'config.json')))

    if isinstance(filename_db,type(None) ): 
        filename_db = os.path.join(__refdata__, 'opacities', inputs['opacities']['files']['opacity'])
        if not os.path.isfile(filename_db):
            raise Exception('The opacity file does not exist: '  + filename_db+' The default is to a file opacities.db in reference/opacity/. If you have an older version of PICASO your file might be called opacity.db. Consider just adding the correct path to filename_db=')
    elif not isinstance(filename_db,type(None) ): 
        if not os.path.isfile(filename_db):
            raise Exception('The opacity file youve entered does not exist: '  + filename_db)

           
    if isinstance(raman_db,type(None) ): raman_db = os.path.join(__refdata__, 'opacities', inputs['opacities']['files']['raman'])

    if resample != 1:
        print("YOU ARE REQUESTING RESAMPLING!!")

    opacityclass=RetrieveOpacities(
                filename_db, 
                raman_db,
                wave_range = wave_range, resample = resample)
    return opacityclass

class inputs():
    """Class to setup planet to run

    Parameters
    ----------
    calculation: str 
        (Optional) Controls planet or brown dwarf calculation. Default = 'planet'. Other option is "browndwarf".
    chemeq : bool 
        (Optional) If true, loads in full chemical equilibrium grid, which will take a little bit of time. 
        Default = False

    Attributes
    ----------
    phase_angle() : set phase angle
    gravity() : set gravity
    star() : set stellar spec
    atmosphere() : set atmosphere composition and PT
    clouds() : set cloud profile
    approx()  : set approximation
    spectrum() : create spectrum
    """
    def __init__(self, calculation='planet', chemeq=False):

        self.inputs = json.load(open(os.path.join(__refdata__,'config.json')))

        if 'brown' in calculation:
            self.nostar()

        #if runnng chemical equilibrium, need to load chemeq grid
        if chemeq: self.chemeq_pic = pk.load(open(os.path.join(__refdata__,'chemistry','chem_full.pic'),'rb'), encoding='latin1')


    def phase_angle(self, phase=0,num_gangle=10, num_tangle=1,symmetry=False):
        """Define phase angle and number of gauss and tchebychev angles to compute. 
        Controls all geometry of the calculation. Computes latitude and longitudes. 
        Computes cos theta and incoming and outgoing angles. Adds everything to class. 

        Please see geometry notebook for a deep dive into how these angles are incorporated 
        into the calculation.

        Typical numbers:
        - 1D Thermal: num_gangle = 10, num_tangle=1
        - 1D Reflected light with zero phase : num_gangle = 10, num_tangle=1
        - 1D Reflected light with non zero phase : num_gangle = 8, num_tangle=8
        - 3D Thermal or Reflected : angles will depend on 3D features interested in resolving.
        
        Parameters
        ----------
        phase : float,int
            Phase angle in radians 
        num_gangle : int 
            Number of Gauss angles to integrate over facets (Default is 10). This 
            is defined as angles over the full sphere. So 10 as the default compute the RT 
            with 5 points on the west hemisphere and 5 points on the west. 
            Higher numbers will slow down code. 
        num_tangle : int 
            Number of Tchebyshev angles to integrate over facets (Default is 1, which automatically 
            assumes symmetry). 
            Must be even if considering symmetry. 
        symmetry : bool 
            Default is False. Note, if num_tangle=1 you are automatically considering symmetry.
            This will only compute the unique incoming angles so that the calculation 
            isn't doing repetetive models. E.g. if num_tangles=10, num_gangles=10. Instead of running a 10x10=100
            FT calculations, it will only run 5x5 = 25 (or a quarter of the sphere).
        """
        if ((num_tangle==1) or (num_gangle==1)): 
            #this is here so that we can compare to older models 
            #this model only works for full phase calculations

            if phase!=0: raise Exception('This method is faster because it makes use of symmetry \
                and only computs one fourth of the sphere. ')
            if num_gangle==1:  raise Exception('Please resubmit your run with num_tangle=1. \
                Chebyshev angles are used for 3d implementation only.')

            num_gangle = int(num_gangle/2) #utilizing symmetry

            #currently only supported values
            possible_values = np.array([5,6,7,8])
            idx = (np.abs(possible_values - num_gangle)).argmin() 
            num_gangle = possible_values[idx]


            gangle, gweight, tangle, tweight = get_angles_1d(num_gangle) 

            ng = len(gangle)
            nt = len(tangle)

            ubar0, ubar1, cos_theta,lat,lon = compute_disco(ng, nt, gangle, tangle, phase)

            geom = {}
            #build dictionary
            geom['num_gangle'], geom['num_tangle'] = ng, nt 
            geom['gangle'], geom['gweight'],geom['tangle'], geom['tweight'] = gangle,gweight,tangle,tweight
            geom['latitude'], geom['longitude']  = lat, lon 
            geom['cos_theta'] = 1.0 
            geom['ubar0'], geom['ubar1'] = ubar0, ubar1 
            self.inputs['phase_angle'] = phase
            self.inputs['disco'] = geom
        else: 

            #this is the new unhardcoded way except 
            self.inputs['phase_angle'] = phase
            
            ng = int(num_gangle)
            nt = int(num_tangle)

            gangle,gweight,tangle,tweight = get_angles_3d(ng, nt) 

            unique_geom = {}
            full_geom = {}

            #planet disk is divided into gaussian and chebyshev angles and weights for perfoming the 
            #intensity as a function of planetary phase angle 
            ubar0, ubar1, cos_theta,lat,lon = compute_disco(ng, nt, gangle, tangle, phase)

            #build dictionary
            full_geom['num_gangle'] = ng
            full_geom['num_tangle'] = nt 
            full_geom['gangle'], full_geom['gweight'],full_geom['tangle'], full_geom['tweight'] = gangle,gweight,tangle,tweight
            full_geom['latitude'], full_geom['longitude']  = lat, lon 
            full_geom['cos_theta'] = cos_theta 
            full_geom['ubar0'], full_geom['ubar1'] = ubar0, ubar1 

            #if running symmetric calculations
            if symmetry:
                if (phase!=0): 
                    raise Exception('If phase is non zero then you cannot utilize symmetry to \
                        reduce computation speed.')
                if (((num_gangle==2) | (num_tangle==2)) and symmetry):
                    raise Exception('Youve selected num_tangle or num_gangle of 2 \
                        however for symmetry to be utilized we need at LEAST two points \
                        on either side of the symmetric axis (e.g. num angles >=4 or 1).\
                        Conducting a help(phase_angle) will tell you \
                        Typical num_tangle and num_gangles to use')
                if ((np.mod(num_tangle,2) !=0 ) and (num_tangle !=1 )):
                    raise Exception('Youve selected an odd num_tangle that isnt 1 \
                        however for symmetry to be utilized we need at LEAST two points \
                        on either side of the symmetric axis (e.g. num angles >=4 or 1).\
                        Conducting a help(phase_angle) will tell you \
                        Typical num_tangle and num_gangles to use')
                if ((np.mod(num_gangle,2) !=0 ) and (num_gangle !=1 )):
                    raise Exception('Youve selected an odd num_gangle that isnt 1 \
                        however for symmetry to be utilized we need at LEAST two points \
                        on either side of the symmetric axis (e.g. num angles >=4 or 1).\
                        Conducting a help(phase_angle) will tell you \
                        Typical num_tangle and num_gangles to use')

                unique_geom['symmetry'] = 'true'
                nt_uni = len(np.unique((tweight*1e6).astype(int))) #unique to 1ppm  
                unique_geom['num_tangle'] = nt_uni
                ng_uni = len(np.unique((gweight*1e6).astype(int)))  #unique to 1ppm 
                unique_geom['num_gangle'] = ng_uni
                unique_geom['ubar1'] = ubar1[0:ng_uni, 0:nt_uni]
                unique_geom['ubar0'] = ubar0[0:ng_uni, 0:nt_uni]
                unique_geom['latitude'] = lat[0:nt_uni]
                unique_geom['longitude'] = lon[0:ng_uni] 
                #adjust weights since we will have to multiply everything by 2 
                adjust_gweight = num_tangle/nt_uni
                adjust_tweight = num_gangle/ng_uni
                unique_geom['gangle'], unique_geom['gweight'] = gangle[0:ng_uni] ,adjust_gweight*gweight[0:ng_uni]
                unique_geom['tangle'], unique_geom['tweight'] = tangle[0:nt_uni],adjust_tweight*tweight[0:nt_uni]
                unique_geom['cos_theta'] = cos_theta 
                #add this to disco
                self.inputs['disco'] = unique_geom
                self.inputs['disco']['full_geometry'] = full_geom
            else: 
                full_geom['symmetry'] = 'false'
                self.inputs['disco'] = full_geom

    def gravity(self, gravity=None, gravity_unit=None, 
        radius=None, radius_unit=None, mass = None, mass_unit=None):
        """
        Get gravity based on mass and radius, or gravity inputs 

        Parameters
        ----------
        gravity : float 
            (Optional) Gravity of planet 
        gravity_unit : astropy.unit
            (Optional) Unit of Gravity
        radius : float 
            (Optional) radius of planet MUST be specified for thermal emission!
        radius_unit : astropy.unit
            (Optional) Unit of radius
        mass : float 
            (Optional) mass of planet 
        mass_unit : astropy.unit
            (Optional) Unit of mass 
        """
        if (mass is not None) and (radius is not None):
            m = (mass*mass_unit).to(u.g)
            r = (radius*radius_unit).to(u.cm)
            g = (c.G.cgs * m /  (r**2)).value
            self.inputs['planet']['radius'] = r.value
            self.inputs['planet']['radius_unit'] = 'cm'
            self.inputs['planet']['mass'] = m.value
            self.inputs['planet']['mass_unit'] = 'g'
            self.inputs['planet']['gravity'] = g
            self.inputs['planet']['gravity_unit'] = 'cm/(s**2)'
        elif gravity is not None:
            g = (gravity*gravity_unit).to('cm/(s**2)')
            g = g.value
            self.inputs['planet']['gravity'] = g
            self.inputs['planet']['gravity_unit'] = 'cm/(s**2)'
            self.inputs['planet']['radius'] = np.nan
            self.inputs['planet']['radius_unit'] = 'Radius not specified'
            self.inputs['planet']['mass'] = np.nan
            self.inputs['planet']['mass_unit'] = 'Mass not specified'
        else: 
            raise Exception('Need to specify gravity or radius and mass + additional units')

    def nostar(self):
        """
        Turns off planet specific things, so program can run as usual
        """
        self.inputs['approx']['raman'] = 2 #turning off raman scattering
        self.inputs['star']['database'] = 'nostar'
        self.inputs['star']['temp'] = 'nostar'
        self.inputs['star']['logg'] = 'nostar'
        self.inputs['star']['metal'] = 'nostar'
        self.inputs['star']['radius'] = 'nostar'
        self.inputs['star']['radius_unit'] = 'nostar' 
        self.inputs['star']['flux'] = 'nostar' 
        self.inputs['star']['wno'] = 'nostar' 
        self.inputs['star']['semi_major'] = 'nostar' 
        self.inputs['star']['semi_major_unit'] = 'nostar' 

    def star(self, opannection,temp=None, metal=None, logg=None ,radius = None, radius_unit=None,
        semi_major=None, semi_major_unit = None,
        database='ck04models',filename=None, w_unit=None, f_unit=None):
        """
        Get the stellar spectrum using pysynphot and interpolate onto a much finer grid than the 
        planet grid. 

        Parameters
        ----------
        opannection : class picaso.RetrieveOpacities
            This is the opacity class and it's needed to get the correct wave info and raman scattering cross sections
        temp : float 
            (Optional) Teff of the stellar model if using the stellar database feature. 
            Not needed for filename option. 
        metal : float 
            (Optional) Metallicity of the stellar model if using the stellar database feature. 
            Not needed for filename option. 
        logg : float 
            (Optional) Logg cgs of the stellar model if using the stellar database feature. 
            Not needed for filename option. 
        radius : float 
            (Optional) Radius of the star. Only needed as input if you want relative flux units (Fp/Fs)
        radius_unit : astropy.unit
            (Optional) Any astropy unit (e.g. `radius_unit=astropy.unit.Unit("R_sun")`)
        semi_major : float 
            (Optional) Semi major axis of the planet. Only needed to compute fp/fs for albedo calculations. 
        semi_major_unit : astropy.unit 
            (Optional) Any astropy unit (e.g. `radius_unit=astropy.unit.Unit("au")`)
        database : str 
            (Optional)The database to pull stellar spectrum from. See documentation for pysynphot. 
        filename : str 
            (Optional) Upload your own stellar spectrum. File format = two column white space (wave, flux). 
            Must specify w_unit and f_unit 
        w_unit : str 
            (Optional) Used for stellar file wave units. Needed for filename input.
            Pick: 'um', 'nm', 'cm', 'hz', or 'Angs'
        f_unit : str 
            (Optional) Used for stellar file flux units. Needed for filename input.
            Pick: 'FLAM' or 'Jy' or 'erg/cm2/s/Hz'
        """
        #most people will just upload their thing from a database
        if (not isinstance(radius, type(None))):
            r = (radius*radius_unit).to(u.cm).value
            radius_unit='cm'
        else :
            r = np.nan
            radius_unit = "Radius not supplied"

        #add semi major axis if supplied 
        if (not isinstance(semi_major, type(None))):
            semi_major = (semi_major*semi_major_unit).to(u.cm).value
            semi_major_unit='cm'
        else :
            semi_major = np.nan
            semi_major_unit = "Semi Major axis not supplied"        

        #upload from file  
        if (not isinstance(filename,type(None))):
            star = np.genfromtxt(filename, dtype=(float, float), names='w, f')
            flux = star['f']
            wave = star['w']
            #sort if not in ascending order 
            sort = np.array([wave,flux]).T
            sort= sort[sort[:,0].argsort()]
            wave = sort[:,0]
            flux = sort[:,1] 
            if w_unit == 'um':
                WAVEUNITS = 'um' 
            elif w_unit == 'nm':
                WAVEUNITS = 'nm'
            elif w_unit == 'cm' :
                WAVEUNITS = 'cm'
            elif w_unit == 'Angs' :
                WAVEUNITS = 'angstrom'
            elif w_unit == 'Hz' :
                WAVEUNITS = 'Hz'
            else: 
                raise Exception('Stellar units are not correct. Pick um, nm, cm, hz, or Angs')        

            #http://www.gemini.edu/sciops/instruments/integration-time-calculators/itc-help/source-definition
            if f_unit == 'Jy':
                FLUXUNITS = 'jy' 
            elif f_unit == 'FLAM' :
                FLUXUNITS = 'FLAM'
            elif f_unit == 'erg/cm2/s/Hz':
                flux = flux*1e23
                FLUXUNITS = 'jy' 
            else: 
                raise Exception('Stellar units are not correct. Pick FLAM or Jy or erg/cm2/s/Hz')

            sp = psyn.ArraySpectrum(wave, flux, waveunits=WAVEUNITS, fluxunits=FLUXUNITS)        #Convert evrything to nanometer for converstion based on gemini.edu  
            sp.convert("um")
            sp.convert('flam') #ergs/cm2/s/ang
            wno_star = 1e4/sp.wave[::-1] #convert to wave number and flip
            flux_star = sp.flux[::-1]*1e8 #flip and convert to ergs/cm3/s here to get correct order         
        

        elif ((not isinstance(temp, type(None))) & (not isinstance(metal, type(None))) & (not isinstance(logg, type(None)))):
            sp = psyn.Icat(database, temp, metal, logg)
            sp.convert("um")
            sp.convert('flam') 
            wno_star = 1e4/sp.wave[::-1] #convert to wave number and flip
            flux_star = sp.flux[::-1]*1e8    #flip here and convert to ergs/cm3/s to get correct order
        else: 
            raise Exception("Must enter 1) filename,w_unit & f_unit OR 2)temp, metal & logg ")

        wno_planet = opannection.wno
        max_shift = np.max(wno_planet)+6000 #this 6000 is just the max raman shift we could have 
        min_shift = np.min(wno_planet) -2000 #it is just to make sure we cut off the right wave ranges

        #this adds stellar shifts 'self.raman_stellar_shifts' to the opacity class
        #the cross sections are computed later 
        if self.inputs['approx']['raman'] == 0: 
            #do a fail safe to make sure that star is on a fine enough grid for planet case 
            fine_wno_star = np.linspace(min_shift, max_shift, len(wno_planet)*5)
            fine_flux_star = np.interp(fine_wno_star,wno_star, flux_star)
            opannection.compute_stellar_shits(fine_wno_star, fine_flux_star)
        else :
            fine_wno_star = wno_planet
            fine_flux_star = np.interp(wno_planet,wno_star, flux_star)  
            opannection.unshifted_stellar_spec =fine_flux_star

        self.inputs['star']['database'] = database
        self.inputs['star']['temp'] = temp
        self.inputs['star']['logg'] = logg
        self.inputs['star']['metal'] = metal
        self.inputs['star']['radius'] = r 
        self.inputs['star']['radius_unit'] = radius_unit 
        self.inputs['star']['flux'] = fine_flux_star 
        self.inputs['star']['wno'] = fine_wno_star 
        self.inputs['star']['semi_major'] = semi_major 
        self.inputs['star']['semi_major_unit'] = semi_major_unit 

    def atmosphere(self, df=None, filename=None, exclude_mol=None, **pd_kwargs):
        """
        Builds a dataframe and makes sure that minimum necessary parameters have been suplied.
        Sets number of layers in model.  

        Parameters
        ----------
        df : pandas.DataFrame or dict
            (Optional) Dataframe with volume mixing ratios and pressure, temperature profile. 
            Must contain pressure (bars) at least one molecule
        filename : str 
            (Optional) Filename with pressure, temperature and volume mixing ratios.
            Must contain pressure at least one molecule
        exclude_mol : list of str 
            (Optional) List of molecules to ignore from file
        pd_kwargs : kwargs 
            Key word arguments for pd.read_csv to read in supplied atmosphere file 
        """

        if not isinstance(df, type(None)):
            if ((not isinstance(df, dict )) & (not isinstance(df, pd.core.frame.DataFrame ))): 
                raise Exception("df must be pandas DataFrame or dictionary")
            else:
                self.nlevel=df.shape[0] 
        elif not isinstance(filename, type(None)):
            df = pd.read_csv(filename, **pd_kwargs)
            self.nlevel=df.shape[0] 
        #elif not isinstance(pt_params, type(None)): 
        #    self.inputs['atmosphere']['pt_params'] = pt_params
        #    self.nlevel=61 #default n of levels for parameterization
        #    raise Exception('Pt parameterization not in yet')

        if 'pressure' not in df.keys(): 
            raise Exception("Check column names. `pressure` must be included.")

        if ('temperature' not in df.keys()):
            raise Exception("`temperature` not specified as a column/key name")

        if not isinstance(exclude_mol, type(None)):
            df = df.drop(exclude_mol, axis=1)
            self.inputs['atmosphere']['exclude_mol'] = exclude_mol

        self.inputs['atmosphere']['profile'] = df.sort_values('pressure').reset_index(drop=True)

        #lastly check to see if the atmosphere is non-H2 dominant. 
        #if it is, let's turn off Raman scattering for the user. 
        if (("H2" not in df.keys()) and (self.inputs['approx']['raman'] != 2)):
            print("Turning off Raman for Non-H2 atmosphere")
            self.approx(raman='none')
        elif (("H2" in df.keys()) and (self.inputs['approx']['raman'] != 2)): 
            if df['H2'].min() < 0.7: 
                print("Turning off Raman for Non-H2 atmosphere")
                self.approx(raman='none')

    def sonora(self, sonora_path, teff, chem='low'):
        """
        This queries Sonora temperature profile that can be downloaded from profiles.tar on 
        Zenodo: [profile.tar file](https://zenodo.org/record/1309035#.Xo5GbZNKjGJ)

        Note gravity is not an input because it grabs gravity from self. 

        Parameters
        ----------
        sonora_path : str   
            Path to the untarred profile.tar file from sonora grid 
        teff : float 
            teff to query. does not have to be exact (it will query one the nearest neighbor)
        chem : str 
            Default = 'low'. There are two sonora chemistry grids. the default is use low. 
            There is sublety to this that is going to be explained at length in the sonora grid paper. 
            Until then, ONLY use chem='low' unless you have reached out to one of the developers. 
        """
        try: 
            g = self.inputs['planet']['gravity']/100 #m/s2 for sonora
        except AttributeError : 
            raise Exception('Oops! Looks like gravity has not been set. Can you please \
                run the gravity function to set gravity')

        flist = os.listdir(os.path.join(sonora_path))
        flist = [i.split('/')[-1] for i in flist if 'gz' in i]
        ts = [i.split('g')[0][1:] for i in flist if 'gz' in i]
        gs = [i.split('g')[1].split('nc')[0] for i in flist]

        pairs = [[ind, float(i),float(j)] for ind, i, j in zip(range(len(ts)), ts, gs)]
        coordinate = [teff, g]

        get_ind = min(pairs, key=lambda c: math.hypot(c[1]- coordinate[0], c[2]-coordinate[1]))[0]

        build_filename = 't'+ts[get_ind]+'g'+gs[get_ind]+'nc_m0.0.cmp.gz'
        ptchem = pd.read_csv(os.path.join(sonora_path,build_filename),delim_whitespace=True,compression='gzip')
        ptchem = ptchem.rename(columns={'P(BARS)':'pressure',
                                        'TEMP':'temperature',
                                        'HE':'He'})

        self.inputs['atmosphere']['profile'] = ptchem.loc[:,['pressure','temperature']]
        if chem == 'high':
            self.channon_grid_high(filename=os.path.join(__refdata__, 'chemistry','grid75_feh+000_co_100_highP.txt'))
        elif chem == 'low':
            self.channon_grid_low(filename=os.path.join(__refdata__,'chemistry','visscher_abunds_m+0.0_co1.0' ))
        self.inputs['atmosphere']['sonora_filename'] = build_filename
        self.nlevel = ptchem.shape[0]

    def chemeq(self, CtoO, Met):
        """
        This interpolates from a precomputed grid of CEA runs (run by M.R. Line)

        Parameters
        ----------
        CtoO : float
            C to O ratio (solar = 0.55)
        Met : float 
            Metallicity relative to solar (solar = 1)
        """
         
        P, T = self.inputs['atmosphere']['profile']['pressure'].values,self.inputs['atmosphere']['profile']['temperature'].values
        
        T[T<400] = 400
        T[T>2800] = 2800

        logCtoO, logMet, Tarr, logParr, gases=self.chemeq_pic
        assert Met <= 10**np.max(logMet), 'Metallicity entered is higher than the max of the grid: M/H = '+ str(np.max(10**logMet))+'. Make sure units are not in log. Solar M/H = 1.'
        assert CtoO <= 10**np.max(logCtoO), 'C/O ratio entered is higher than the max of the grid: C/O = '+ str(np.max(10**logCtoO))+'. Make sure units are not in log. Solar C/O = 0.55'
        assert Met >= 10**np.min(logMet), 'Metallicity entered is lower than the min of the grid: M/H = '+ str(np.min(10**logMet))+'. Make sure units are not in log. Solar M/H = 1.'
        assert CtoO >= 10**np.min(logCtoO), 'C/O ratio entered is lower than the min of the grid: C/O = '+ str(np.min(10**logCtoO))+'. Make sure units are not in log. Solar C/O = 0.55'

        loggas=np.log10(gases)
        Ngas = loggas.shape[3]
        gas=np.zeros((Ngas,len(P)))
        for j in range(Ngas):
            gas_to_interp=loggas[:,:,:,j,:]
            IF=RegularGridInterpolator((logCtoO, logMet, np.log10(Tarr),logParr),gas_to_interp,bounds_error=False)
            for i in range(len(P)):
                gas[j,i]=10**IF(np.array([np.log10(CtoO), np.log10(Met), np.log10(T[i]), np.log10(P[i])]))
        H2Oarr, CH4arr, COarr, CO2arr, NH3arr, N2arr, HCNarr, H2Sarr,PH3arr, C2H2arr, C2H6arr, Naarr, Karr, TiOarr, VOarr, FeHarr, Harr,H2arr, Hearr, mmw=gas

        df = pd.DataFrame({'H2O': H2Oarr, 'CH4': CH4arr, 'CO': COarr, 'CO2': CO2arr, 'NH3': NH3arr, 
                           'N2' : N2arr, 'HCN': HCNarr, 'H2S': H2Sarr, 'PH3': PH3arr, 'C2H2': C2H2arr, 
                           'C2H6' :C2H6arr, 'Na' : Naarr, 'K' : Karr, 'TiO': TiOarr, 'VO' : VOarr, 
                           'Fe': FeHarr,  'H': Harr, 'H2' : H2arr, 'He' : Hearr, 'temperature':T, 
                           'pressure': P})
        self.inputs['atmosphere']['profile'] = df
        return 

    def channon_grid_high(self,filename=None):
        if isinstance(filename, type(None)):filename=os.path.join(__refdata__,'chemistry','grid75_feh+000_co_100_highP.txt')
        #df = self.inputs['atmosphere']['profile']
        df = self.inputs['atmosphere']['profile'].sort_values('pressure').reset_index(drop=True)
        self.nlevel = df.shape[0]
        
        player = df['pressure'].values
        tlayer  = df['temperature'].values
        
        channon = pd.read_csv(filename,delim_whitespace=True)
        #get molecules list from channon db
        mols = list(channon.keys())
        mols.pop(mols.index('pressure'))
        mols.pop(mols.index('temperature'))
        
        #add molecules to df
        df = df.reindex(columns = ['pressure','temperature']+mols) 

        #add pt_ids so we can grab the right points
        pt_pairs = channon.loc[:,['pressure','temperature']]
        pt_pairs['pt_id'] = np.arange(pt_pairs.shape[0])
        channon['pt_id'] = pt_pairs['pt_id']

        pt_pairs = pt_pairs.values#p,t,id
        
        #find index corresponding to each pair
        ind_pt=[min(pt_pairs, 
                    key=lambda c: math.hypot(c[0]- coordinate[0], c[1]-coordinate[1]))[2] 
                        for coordinate in  zip(np.log10(player),tlayer)]
        #build dataframe with chemistry
        for i in range(df.shape[0]):
            pair_for_layer = ind_pt[i]
            df.loc[i,mols] = channon.loc[channon['pt_id'] == pair_for_layer,mols].values
        df['ptid'] = ind_pt
        
        self.inputs['atmosphere']['profile'] = df

    def channon_grid_low(self, filename = None, interp_window = 11, interp_poly=2):
        """
        Interpolate from visscher grid
        """
        if isinstance(filename, type(None)):filename= os.path.join(__refdata__,'chemistry','visscher_abunds_m+0.0_co1.0')
        a = pd.read_csv(filename)
        mols = list(a.loc[:, ~a.columns.isin(['Unnamed: 0','pressure','temperature'])].keys())
        #get pt from what users have already input
        player_tlayer = self.inputs['atmosphere']['profile'].loc[:,['pressure','temperature']].sort_values('pressure').reset_index(drop=True)
        self.nlevel = player_tlayer.shape[0]

        #loop through all pressure and temperature layers
        for i in player_tlayer.index:
            p,t = player_tlayer.loc[i,['pressure','temperature']]
            #at each layer compute the distance to each point in the visscher table
            #create a weight that is 1/d**2
            a['weight'] = (1/((np.log10(a['pressure']) - np.log10(p))**2 + 
                                       (a['temperature'] - t)**2))
            #sort by the weight and only pick the TWO nearest neighbors. 
            #Note, there was a sensitivity study to determine how many nearest neighbors to choose 
            #It was determined that no more than 2 is best. 
            w = a.sort_values(by='weight',ascending=False).iloc[0:2,:]
            #now loop through all the molecules
            for im in mols:
                #compute the weighted mean of the mixing ratio
                weighted_mean = (np.sum(w['weight']*(w.loc[:,im]))/np.sum(w['weight']))
                player_tlayer.loc[i,im] = weighted_mean
        #now pass through everything and use savgol filter to average 
        for im in mols:
            y = np.log10(player_tlayer.loc[:,im])
            s = savgol_filter(y , interp_window, interp_poly)
            player_tlayer.loc[:,im] = 10**s #

        self.inputs['atmosphere']['profile'] = player_tlayer

    def guillot_pt(self, Teq, T_int=100, logg1=-1, logKir=-1.5, alpha=0.5,nlevel=61, p_bottom = 1.5, p_top = -6):
        """
        Creates temperature pressure profile given parameterization in Guillot 2010 TP profile
        called in fx()

        Parameters
        ----------
        Teq : float 
            equilibrium temperature 
        T_int : float 
            Internal temperature, if low (100) currently set to 100 for everything  
        kv1 : float 
            see parameterization Guillot 2010 (10.**(logg1+logKir))
        kv2 : float
            see parameterization Guillot 2010 (10.**(logg1+logKir))
        kth : float
            see parameterization Guillot 2010 (10.**logKir)
        alpha : float , optional
            set to 0.5
        nlevel : int, optional
            Number of atmospheric layers
        p_bottom : float, optional 
            Log pressure (bars) of the lower bound pressure 
        p_top : float , optional
            Log pressure (bars) of the TOA 
            
        Returns
        -------
        T : numpy.array 
            Temperature grid 
        P : numpy.array
            Pressure grid
                
        """
        kv1, kv2 =10.**(logg1+logKir),10.**(logg1+logKir)
        kth=10.**logKir

        Teff = T_int
        f = 1.0  # solar re-radiation factor
        A = 0.0  # planetary albedo
        g0 = self.inputs['planet']['gravity']/100.0 #cm/s2 to m/s2

        # Compute equilibrium temperature and set up gamma's
        T0 = Teq
        gamma1 = kv1/kth #Eqn. 25
        gamma2 = kv2/kth

        # Initialize arrays
        logtau =np.arange(-10,20,.1)
        tau =10**logtau

        #computing temperature
        T4ir = 0.75*(Teff**(4.))*(tau+(2.0/3.0))
        f1 = 2.0/3.0 + 2.0/(3.0*gamma1)*(1.+(gamma1*tau/2.0-1.0)*sp.exp(-gamma1*tau))+2.0*gamma1/3.0*(1.0-tau**2.0/2.0)*special.expn(2.0,gamma1*tau)
        f2 = 2.0/3.0 + 2.0/(3.0*gamma2)*(1.+(gamma2*tau/2.0-1.0)*sp.exp(-gamma2*tau))+2.0*gamma2/3.0*(1.0-tau**2.0/2.0)*special.expn(2.0,gamma2*tau)
        T4v1=f*0.75*T0**4.0*(1.0-alpha)*f1
        T4v2=f*0.75*T0**4.0*alpha*f2
        T=(T4ir+T4v1+T4v2)**(0.25)
        P=tau*g0/(kth*0.1)/1.E5
        self.nlevel=nlevel 
        logP = np.linspace(p_top,p_bottom,nlevel)
        newP = 10.0**logP
        T = np.interp(logP,np.log10(P),T)

        self.inputs['atmosphere']['profile']  = pd.DataFrame({'temperature': T, 'pressure':newP})

        # Return TP profile
        return self.inputs['atmosphere']['profile'] 

    def TP_line_earth(self,P,Tsfc=294.0, Psfc=1.0, gam_trop=0.18, Ptrop=0.199, 
        gam_strat=-0.045,Pstrat=0.001,nlevel=150):
        """
        Author: Mike R. Line 

        Estimates Earth's pressure-temperature profile. All default 
        values have been tuned to semi reproduce Earth's temperature 
        pressure profile.

        Parameters
        ----------
        P : array 
            Pressure array usually np.logspace(-6,2,nlevel)
        Tsfc : float,optional 
            Surface Temperature (K). Earth is 294 K 
        Psfc : float ,optional
            Surface Pressure (bar). Earth is 1 bar. 
        gam_trop : float ,optional
            Tropospheric dry lapse rate. Earth is ~0.18 
        Ptrop : float ,optional
            Tropospheric pressure. Earth 0.199 bar. 
        gam_strat : float ,optional
            Stratospheric lapse rate. Earth is -0.045
        Pstrat : float ,optional
            Stratospheric pressure (bars). Earth is 0.001 bar. /
        nlevel : int ,optional
            Number of grid levels 

        Returns 
        -------
        array 
            Temperature array (K)Psfc
        """
        #P = np.logspace(np.log10(1e-6), np.log10(100),nlevel)

        if Ptrop <= P.min(): Ptrop=P.min()
        if Pstrat <= P.min(): Pstrat=P.min()

        T=np.zeros(len(P))

        #troposphere T--adibat

        Ttrop=Tsfc*(P/Psfc)**gam_trop  #P0=sfc p, Trop T
            
        #stratosphere
        Tpause=Ttrop[P <= Ptrop ][-1]  #tropopause Temp
        PPtrop=P[P <= Ptrop ][-1]
        Tstrat=Tpause*(P/PPtrop)**gam_strat

        #merging troposphere and stratosphfere
        T[P >= Ptrop]=Ttrop[P >= Ptrop]
        T[P <= Ptrop]=Tstrat[P <= Ptrop]

        #isothermal below surface (making surface a blackbody)
        T[P >= Psfc]=T[P >= Psfc][0]

        #isothermal above "stratopause" pressure, Pstrat
        T[P<=Pstrat]=T[P<=Pstrat][-1]
        T[T<=10]=10
        T[T>=1000]=1000
         
        self.inputs['atmosphere']['profile']  = pd.DataFrame({'temperature': T, 'pressure':P})

        # Return TP profile
        return self.inputs['atmosphere']['profile'] 


    def atmosphere_3d(self, dictionary=None, filename=None):
        """
        Builds a dataframe and makes sure that minimum necessary parameters have been suplied. 

        Parameters
        ----------
        df : dict
            (Optional) Dataframe with volume mixing ratios and pressure, temperature profile. 
            Must contain pressure (bars) and at least one molecule
        filename : str 
            (Optional) Pickle filename that is a dictionary structured: 
            dictionary[int(lat) in degrees][int(lon) in degrees] = pd.DataFrame({'pressure':[], 'temperature': [],
            'H2O':[]....}). As well as df['phase_angle'] = 0 (in radians).
            Therefore, the latitude and longitude keys should be INTEGERS taken from the lat/longs calculated in 
            inputs.phase_angle! Any other meta data should be  represented by a string.
            Must contain pressure and at least one molecule

        Examples
        --------

        >>import picaso.justdoit as jdi
        >>new = jdi.inputs()
        >>new.phase_angle(0) #loads the geometry you need to get your 3d model on
        >>lats= new.inputs['disco']['latitude']*180/np.pi).astype(int)
        >>>lons = (new.inputs['disco']['longitude']*180/np.pi).astype(int)

        Build an empty dictionary to be filled later 

        >>dict_3d = {la: {lo : [] for lo in lons} for la in lats} #build empty one 
        >>dict_3d['phase_angle'] = 0 #this is a double check for consistency 
        
        Now you need to bin your GCM output on this grid and fill the dictionary with the necessary dataframes. For example:
        
        >>dict_3d[lats[0]][lons[0]] = pd.DataFrame({'pressure':[], 'temperature': [],'H2O':[]})

        Once this is filled you can add it to this function 

        >>new.atmosphere_3d(dictionary=dict_3d)
        """

        if (isinstance(dictionary, dict) and isinstance(filename, type(None))):
            df3d = dictionary

        elif (isinstance(dictionary, type(None)) and isinstance(filename, str)):
            df3d = pk.load(open(filename,'rb'))
        else:
            raise Exception("Please input either a dict using dictionary or a str using filename")

        #get lats and lons out of dictionary and puts them in reverse 
        #order so that they match what comes out of disco calcualtion
        lats = np.sort(list(i for i in df3d.keys() if isinstance(i,int))) #latitude is positive to negative 
        lons = np.sort(list(i for i in df3d[lats[0]].keys() if isinstance(i,int))) #longitude is negative to positive 
        #check they are about the same as the ones computed in phase angle 
        df3d_nt = len(lats)
        df3d_ng =  len(df3d[lats[0]].keys())

        assert self.inputs['disco']['num_gangle'] == int(df3d_nt), 'Number of gauss angles input does not match creation of 3d input file. Check function `inputs.phase_angle()`. num_gangle=10 is set by default and you may have to change it.'
        assert self.inputs['disco']['num_tangle']  == int(df3d_ng), 'Number of Tchebyshev angles does not match creation of 3d input file.  Check function `inputs.phase_angle()`.  num_tangle=10 is set by default and you may have to change it.'
        
        self.nlevel=df3d[lats[0]][lons[0]].shape[0]

        #now check that the lats and longs are about the same
        for ilat ,nlat in  zip(np.sort(self.inputs['disco']['latitude']*180/np.pi), lats): 
            np.testing.assert_almost_equal(int(ilat) ,nlat, decimal=0,err_msg='Latitudes from dictionary(units degrees) are not the same as latitudes computed in inputs.phase_angle', verbose=True)

        for ilon ,nlon in  zip(np.sort(self.inputs['disco']['longitude']*180/np.pi), lons): 
            np.testing.assert_almost_equal(int(ilon) ,nlon , decimal=0,err_msg='Longitudes from dictionary(units degrees) are not the same as longitudes computed in inputs.phase_angle', verbose=True)
        
        self.inputs['atmosphere']['profile'] = df3d

    def surface_reflect(self, albedo, wavenumber, old_wavenumber = None):
        """
        Set atmospheric surface reflectivity

        Parameters
        ----------
        albedo : float
            Set constant albedo for surface reflectivity 
        """
        if isinstance(albedo, (float, int)):
            self.inputs['surface_reflect'] = np.array([albedo]*len(wavenumber))
        elif isinstance(albedo, (list, np.ndarray)): 
            if isinstance(old_wavenumber, type(None)):
                self.inputs['surface_reflect'] = albedo
            else: 
                self.inputs['surface_reflect'] = np.interp(wavenumber, old_wavenumber, albedo)
    
    def clouds_reset(self):
        """Reset cloud dict to zeros"""
        df = self.inputs['clouds']['profile']
        zeros=np.zeros(196*(self.nlevel-1))

        #add in cloud layers 
        df['g0'] = zeros
        df['w0'] = zeros
        df['opd'] = zeros
        self.inputs['clouds']['profile'] = df
        
    def clouds(self, filename = None, g0=None, w0=None, opd=None,p=None, dp=None,df =None,**pd_kwargs):
        """
        Cloud specification for the model. Clouds are parameterized by a single scattering albedo (w0), 
        an assymetry parameter (g0), and a total extinction per layer (opd).

        g0,w0, and opd are both wavelength and pressure dependent. Our cloud models come 
        from eddysed. Their output look something like this where 
        pressure is in bars and wavenumber is inverse cm. We will sort pressure and wavenumber before we reshape
        so the exact order doesn't matter

        pressure wavenumber opd w0 g0
        1.   1.   ... . .
        1.   2.   ... . .
        1.   3.   ... . .
        .     . ... . .
        .     . ... . .
        1.   M.   ... . .
        2.   1.   ... . .
        .     . ... . .
        N.   .  ... . .

        If you are creating your own file you have to make sure that you have a 
        **pressure** (bars) and **wavenumber**(inverse cm) column. We will use this to make sure that your cloud 
        and atmospheric profiles are on the same grid. **If there is no pressure or wavelength parameter
        we will assume that you are on the same grid as your atmospheric input, and on the 
        eddysed wavelength grid! **

        Users can also input their own fixed cloud parameters, by specifying a single value 
        for g0,w0,opd and defining the thickness and location of the cloud. 

        Parameters
        ----------
        filename : str 
            (Optional) Filename with info on the wavelength and pressure-dependent single scattering
            albedo, asymmetry factor, and total extinction per layer. Input associated pd_kwargs 
            so that the resultant output has columns named : `g0`, `w0` and `opd`. If you are not 
            using the eddysed output, you will also need a `wavenumber` and `pressure` column in units 
            of inverse cm, and bars. 
        g0 : float, list of float
            (Optional) Asymmetry factor. Can be a single float for a single cloud. Or a list of floats 
            for two different cloud layers 
        w0 : list of float 
            (Optional) Single Scattering Albedo. Can be a single float for a single cloud. Or a list of floats 
            for two different cloud layers      
        opd : list of float 
            (Optional) Total Extinction in `dp`. Can be a single float for a single cloud. Or a list of floats 
            for two different cloud layers 
        p : list of float 
            (Optional) Bottom location of cloud deck (LOG10 bars). Can be a single float for a single cloud. Or a list of floats 
            for two different cloud layers 
        dp : list of float 
            (Optional) Total thickness cloud deck above p (LOG10 bars). 
            Can be a single float for a single cloud or a list of floats 
            for two different cloud layers 
            Cloud will span 10**(np.log10(p-dp))
        df : pd.DataFrame, dict
            (Optional) Same as what would be included in the file, but in DataFrame or dict form
        """

        #first complete options if user inputs dataframe or dict 
        if (not isinstance(filename, type(None)) & isinstance(df, type(None))) or (isinstance(filename, type(None)) & (not isinstance(df, type(None)))):

            if not isinstance(filename, type(None)):
                df = pd.read_csv(filename, **pd_kwargs)

            cols = df.keys()

            assert 'g0' in cols, "Please make sure g0 is a named column in cld file"
            assert 'w0' in cols, "Please make sure w0 is a named column in cld file"
            assert 'opd' in cols, "Please make sure opd is a named column in cld file"

            #CHECK SIZES

            #if it's a user specified pressure and wavenumber
            if (('pressure' in cols) & ('wavenumber' in cols)):
                df = df.sort_values(['pressure', 'wavenumber']).reset_index(drop=True)
                self.inputs['clouds']['wavenumber'] = df['wavenumber'].unique()
                nwave = len(self.inputs['clouds']['wavenumber'])
                nlayer = len(df['pressure'].unique())
                assert df.shape[0] == (self.nlevel-1)*nwave, "There are {0} rows in the df, which does not equal {1} layers previously specified x {2} wave pts".format(df.shape[0], self.nlevel-1, nwave) 
            
            #if its eddysed, make sure there are 196 wave points 
            else: 
                assert df.shape[0] == (self.nlevel-1)*196, "There are {0} rows in the df, which does not equal {1} layers x 196 eddysed wave pts".format(df.shape[0], self.nlevel-1) 
                
                self.inputs['clouds']['wavenumber'] = get_cld_input_grid('wave_EGP.dat')

            #add it to input
            self.inputs['clouds']['profile'] = df

        #first make sure that all of these have been specified
        elif None in [g0, w0, opd, p,dp]:
            raise Exception("Must either give dataframe/dict, OR a complete set of g0, w0, opd,p,dp to compute cloud profile")
        else:
            pressure_level = self.inputs['atmosphere']['profile']['pressure'].values
            pressure = np.sqrt(pressure_level[1:] * pressure_level[0:-1])#layer

            w = get_cld_input_grid('wave_EGP.dat')

            self.inputs['clouds']['wavenumber'] = w

            pressure_all =[]
            for i in pressure: pressure_all += [i]*len(w)
            wave_all = list(w)*len(pressure)

            df = pd.DataFrame({'pressure':pressure_all,
                                'wavenumber': wave_all })


            zeros=np.zeros(196*(self.nlevel-1))

            #add in cloud layers 
            df['g0'] = zeros
            df['w0'] = zeros
            df['opd'] = zeros
            #loop through all cloud layers and set cloud profile
            for ig, iw, io , ip, idp in zip(g0,w0,opd,p,dp):
                maxp = 10**ip #max pressure is bottom of cloud deck
                minp = 10**(ip-idp) #min pressure 
                df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'g0']= ig
                df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'w0']= iw
                df.loc[((df['pressure'] >= minp) & (df['pressure'] <= maxp)),'opd']= io

            self.inputs['clouds']['profile'] = df

    def virga(self, condensates, directory,
        fsed=1, mh=1, mmw=2.2,kz_min=1e5,sig=2, full_output=False): 
        """
        Runs virga cloud code based on the PT and Kzz profiles 
        that have been added to inptus class.

        Parameters
        ----------
        condensates : str 
            Condensates to run in cloud model 
        fsed : float 
            Sedimentation efficiency 
        mh : float 
            Metallicity 
        mmw : float 
            Atmospheric mean molecular weight  
        """
        
        cloud_p = vj.Atmosphere(condensates,fsed=fsed,mh=mh,
                 mmw = mmw, sig =sig) 

        if 'kz' not in self.inputs['atmosphere']['profile'].keys():
            raise Exception ("Must supply kz to atmosphere/chemistry DataFrame, \
                if running `virga` through `picaso`. This should go in the \
                same place that you specified you pressure-temperature profile. \
                Alternatively, you can manually add it by doing \
                `case.inputs['atmosphere']['profile']['kz'] = KZ`")
        df = self.inputs['atmosphere']['profile'].loc[:,['pressure','temperature','kz']]
        
        cloud_p.gravity(gravity=self.inputs['planet']['gravity'],
                 gravity_unit=u.Unit(self.inputs['planet']['gravity_unit']))#
        
        cloud_p.ptk(df =df, kz_min = kz_min)
        out = vj.compute(cloud_p, as_dict=full_output,
                          directory=directory)

        if not full_output:
            opd, w0, g0 = out
            df = vj.picaso_format(opd, w0, g0)
        else: 
            opd, w0, g0 = out['opd_per_layer'],out['single_scattering'],out['asymmetry']
            df = vj.picaso_format(opd, w0, g0)

        self.clouds(df=df)

        if full_output : return out

    def clouds_3d(self, filename = None, dictionary =None):
        """
        Cloud specification for the model. Clouds are parameterized by a single scattering albedo (w0), 
        an assymetry parameter (g0), and a total extinction per layer (opd).

        g0,w0, and opd are both wavelength and pressure dependent. Our cloud models come 
        from eddysed. Their output looks like this (where level 1=TOA, and wavenumber1=Smallest number)

        level wavenumber opd w0 g0
        1.   1.   ... . .
        1.   2.   ... . .
        1.   3.   ... . .
        .     . ... . .
        .     . ... . .
        1.   M.   ... . .
        2.   1.   ... . .
        .     . ... . .
        N.   .  ... . .

        If you are creating your own file you have to make sure that you have a 
        **pressure** (bars) and **wavenumber**(inverse cm) column. We will use this to make sure that your cloud 
        and atmospheric profiles are on the same grid. **If there is no pressure or wavelength parameter
        we will assume that you are on the same grid as your atmospheric input, and on the 
        eddysed wavelength grid! **

        Users can also input their own fixed cloud parameters, by specifying a single value 
        for g0,w0,opd and defining the thickness and location of the cloud. 

        Parameters
        ----------
        filename : str 
            (Optional) Filename with info on the wavelength and pressure-dependent single scattering
            albedo, asymmetry factor, and total extinction per layer. Input associated pd_kwargs 
            so that the resultant output has columns named : `g0`, `w0` and `opd`. If you are not 
            using the eddysed output, you will also need a `wavenumber` and `pressure` column in units 
            of inverse cm, and bars. 
        dictionary : dict
            (Optional) Same as what would be included in the file, but in dict form

        Examples
        --------

        >>import picaso.justdoit as jdi
        >>new = jdi.inputs()
        >>new.phase_angle(0) #loads the geometry you need to get your 3d model on
        >>lats, lons = int(new.inputs['disco']['latitude']*180/np.pi), int(new.inputs['disco']['longitude']*180/np.pi)

        Build an empty dictionary to be filled later 

        >>dict_3d = {la: {lo : for lo in lons} for la in lats} #build empty one 
        >>dict_3d['phase_angle'] = 0 #add this for a consistensy check
        
        Now you need to bin your 3D clouds GCM output on this grid and fill the dictionary with the necessary dataframes. For example:
        
        >>dict_3d[lats[0]][lons[0]] = pd.DataFrame({'pressure':[], 'temperature': [],'H2O':[]})

        Once this is filled you can add it to this function 

        >>new.clouds_3d(dictionary=dict_3d)
        """

        #first complete options if user inputs dataframe or dict 
        if (isinstance(filename, str) & isinstance(dictionary, type(None))): 
            df = pk.load(open(filename,'rb'))
        elif (isinstance(filename, type(None)) & isinstance(dictionary, dict)):
            df = dictionary
        else: 
            raise Exception("Input must be a filename or a dictionary")

        cols = df[list(df.keys())[0]][list(df[list(df.keys())[0]].keys())[0]].keys()

        assert 'g0' in cols, "Please make sure g0 is a named column in cld file"
        assert 'w0' in cols, "Please make sure w0 is a named column in cld file"
        assert 'opd' in cols, "Please make sure opd is a named column in cld file"
        #again, check lat and lon in comparison to the computed ones 
        #get lats and lons out of dictionary 
        lats = np.sort(list(i for i in df.keys() if isinstance(i,int)))
        lons = np.sort(list(i for i in df[lats[0]].keys() if isinstance(i,int)))    
                
        for ilat ,nlat in  zip(np.sort(self.inputs['disco']['latitude']*180/np.pi), lats): 
            np.testing.assert_almost_equal(int(ilat) ,nlat, decimal=0,err_msg='Latitudes from dictionary(units degrees) are not the same as latitudes computed in inputs.phase_angle', verbose=True)
            
        for ilon ,nlon in  zip(np.sort(self.inputs['disco']['longitude']*180/np.pi), lons): 
            np.testing.assert_almost_equal(int(ilon) ,nlon , decimal=0,err_msg='Latitudes from dictionary(units degrees) are not the same as latitudes computed in inputs.phase_angle', verbose=True)

        #CHECK SIZES
        #if it's a user specified pressure and wavenumber
        if (('pressure' in cols) & ('wavenumber' in cols)):
            for i in df.keys():
                #loop through different lat and longs and make sure that each one of them is ordered correct 
                for j in df[i].keys():
                    df[i][j] = df[i][j].sort_values(['pressure', 'wavenumber']).reset_index(drop=True)
                    self.inputs['clouds']['wavenumber'] = df[i][j]['wavenumber'].unique()
                    nwave = len(self.inputs['clouds']['wavenumber'])
                    nlayer = len(df[i][j]['pressure'].unique())
                    assert df[i][j].shape[0] == (self.nlevel-1)*nwave, "There are {0} rows in the df, which does not equal {1} layers previously specified x {2} wave pts".format(df[i][j].shape[0], self.nlevel-1, nwave) 
                
        #if its eddysed, make sure there are 196 wave points 
        #this does not reorder so it assumes that 
        else: 
            shape = df[list(df.keys())[0]][list(df[list(df.keys())[0]].keys())[0]].shape[0]
            assert shape == (self.nlevel-1)*196, "There are {0} rows in the df, which does not equal {1} layers x 196 eddysed wave pts".format(shape, self.nlevel-1) 
            
            #get wavelength grid from ackerman code
            self.inputs['clouds']['wavenumber'] = get_cld_input_grid('wave_EGP.dat')

            #add it to input
        self.inputs['clouds']['profile'] = df

    def approx(self,single_phase='TTHG_ray',multi_phase='N=2',delta_eddington=True,
        raman='pollack',tthg_frac=[1,-1,2], tthg_back=-0.5, tthg_forward=1,
        p_reference=1,method='Toon', stream=2):
        """
        This function sets all the default approximations in the code. It transforms the string specificatons
        into a number so that they can be used in numba nopython routines. 

        For `str` cases such as `TTHG_ray` users see all the options by using the function `single_phase_options`
        or `multi_phase_options`, etc. 

        single_phase : str 
            Single scattering phase function approximation 
        multi_phase : str 
            Multiple scattering phase function approximation 
        delta_eddington : bool 
            Turns delta-eddington on and off
        raman : str 
            Uses various versions of raman scattering 
        tthg_frac : list 
            Functional of forward to back scattering with the form of polynomial :
            tthg_frac[0] + tthg_frac[1]*g_b^tthg_frac[2]
            See eqn. 6 in picaso paper 
        tthg_back : float 
            Back scattering asymmetry factor gf = g_bar*tthg_back
        tthg_forward : float 
            Forward scattering asymmetry factor gb = g_bar * tthg_forward 
        p_reference : float 
            Reference pressure (bars) This is an arbitrary pressure that 
            corresponds do the user's input of radius. Usually something "at depth"
            around 1-10 bars. 
        method : str
            Toon ('Toon') or spherical harmonics ('SH'). 
        stream : int 
            Two stream or four stream (options are 2 or 4). For 4 stream need to set method='SH'
        """

        self.inputs['approx']['single_phase'] = single_phase_options(printout=False).index(single_phase)
        self.inputs['approx']['multi_phase'] = multi_phase_options(printout=False).index(multi_phase)
        self.inputs['approx']['delta_eddington'] = delta_eddington
        self.inputs['approx']['raman'] =  raman_options().index(raman)
        self.inputs['approx']['method'] = method
        self.inputs['approx']['stream'] = stream
 
        if isinstance(tthg_frac, (list, np.ndarray)):
            if len(tthg_frac) == 3:
                self.inputs['approx']['TTHG_params']['fraction'] = tthg_frac
            else:
                raise Exception('tthg_frac should be of length=3 so that : tthg_frac[0] + tthg_frac[1]*g_b^tthg_frac[2]')
        else: 
            raise Exception('tthg_frac should be a list or ndarray of length=3')

        self.inputs['approx']['TTHG_params']['constant_back'] = tthg_back
        self.inputs['approx']['TTHG_params']['constant_forward']=tthg_forward

        self.inputs['approx']['p_reference']= p_reference

    def spectrum(self, opacityclass, calculation='reflected', dimension = '1d',  full_output=False, 
        plot_opacity= False, as_dict=True):
        """Run Spectrum

        Parameters
        -----------
        opacityclass : class
            Opacity class from `justdoit.opannection`
        calculation : str
            Either 'reflected' or 'thermal' for reflected light or thermal emission. 
            If running a brown dwarf, this will automatically default to thermal    
        dimension : str 
            (Optional) Dimensions of the calculation. Default = '1d'. But '3d' is also accepted. 
            In order to run '3d' calculations, user must build 3d input (see tutorials)
        full_output : bool 
            (Optional) Default = False. Returns atmosphere class, which enables several 
            plotting capabilities. 
        plot_opacity : bool 
            (Optional) Default = False, Creates pop up of the weighted opacity
        as_dict : bool 
            (Optional) Default = True. If true, returns a condensed dictionary to the user. 
            If false, returns the atmosphere class, which can be used for debugging. 
            The class is clunky to navigate so if you are consiering navigating through this, ping one of the 
            developers. 
        """
        try: 
            if self.inputs['star']['radius'] == 'nostar':
                calculation = 'thermal' 
        except KeyError: 
            pass
        
        try:
            a = self.inputs['surface_reflect']
        except KeyError:
            #I don't make people add this as an input so adding a default here if it hasnt
            #been run 
            self.inputs['surface_reflect'] = 0 

            
        return picaso(self, opacityclass,dimension=dimension,calculation=calculation,
            full_output=full_output, plot_opacity=plot_opacity, as_dict=as_dict)

def get_targets():
    """Function to grab available targets

    Returns
    -------
    Dataframe from Exoplanet Archive
    """
    default_query = "https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&format=csv"
    all_planets =  requests.get(default_query)  
    all_planets = all_planets.text.replace(' ','').split('\n')

    # default data doesn't come with all the parameters we need so let's 
    # add those in 
    add_few_elements = '&select='+all_planets[0]+','.join(
        [',pl_trandur','pl_eqt','st_logg','st_metfe','st_j','st_h','st_k&'])

    # use requests to grab the csv formatted data and split by the commas 
    all_planets =  requests.get(default_query.split('&')[0] 
                                + add_few_elements + default_query.split('&')[1] )  
    all_planets = all_planets.text.replace(' ','').split('\n')

    # get into useful pandas dataframe format
    planets_df = pd.DataFrame(columns=all_planets[0].split(','), 
                             data = [i.split(',') for i in all_planets[1:-1]])
    planets_df = planets_df.replace(to_replace='', value=np.nan)

    # convert to float when possible
    for i in planets_df.columns: 
        planets_df[i] = planets_df[i].astype(float,errors='ignore') 

    return planets_df

def load_planet(df, opacity, phase_angle = 0, stellar_db='phoenix', verbose=False,  **planet_kwargs):
    """
    Wrapper to simplify PICASO run. This really turns chimera into a black box. This was created 
    specifically Sagan School tutorial. 

    Parameters
    -----------
    df : pd.DataFrame
        This is single row from `all_planets()`
    opacity : np.array 
        Opacity loaded from opannection
    phase_angle : float 
        Observing phase angle (radians)
    verbose : bool , options
        Print out warnings 
    planet_kwargs : dict 
        List of parameters to supply NexSci database is values don't exist 
    """
    if len(df.index)>1: raise Exception("Dataframe consists of more than 1 row. Make sure to select single planet")

    for i in df.index:

        planet = df.loc[i,:].dropna().to_dict()

        temp = planet.get('st_teff', planet_kwargs.get('st_teff',np.nan))
        if np.isnan(temp) : raise Exception('Stellar temperature is not added to \
            dataframe input or to planet_kwargs through the column/key named st_teff. Please add it to one of them')

        logg = planet.get('st_logg', planet_kwargs.get('st_logg',np.nan))
        if np.isnan(logg) : raise Exception('Stellar logg is not added to \
            dataframe input or to planet_kwargs through the column/key named st_logg. Please add it to one of them')

        logmh = planet.get('st_metfe', planet_kwargs.get('st_metfe',np.nan))
        if np.isnan(logmh) : raise Exception('Stellar Fe/H is not added to \
            dataframe input or to planet_kwargs through the column/key named st_metfe. Please add it to one of them')

        stellar_db = 'phoenix'

        if logmh > 0.5: 
            if verbose: print ('Stellar M/H exceeded max value of 0.5. Value has been reset to the maximum')
            logmh = 0.5
        elif logmh < -4.0 :
            if verbose: print ('Stellar M/H exceeded min value of -4.0 . Value has been reset to the mininum')
            logmh = -4.0 

        if logg > 4.5: 
            if verbose: print ('Stellar logg exceeded max value of 4.5. Value has been reset to the maximum')
            logg = 4.5   


        #the parameters
        #planet/star system params--typically not free parameters in retrieval
        # Planet radius in Jupiter Radii--this will be forced to be 10 bar radius--arbitrary (scaling to this is free par)

        Rp = planet.get('pl_radj', planet_kwargs.get('pl_radj',np.nan))
        if np.isnan(Rp) : raise Exception('Planet Radii is not added to \
            dataframe input or to planet_kwargs through the column/key named pl_radj. J for JUPITER! \
            Please add it to one of them')


        #Stellar Radius in Solar Radii
        Rstar = planet.get('st_rad', planet_kwargs.get('st_rad',np.nan))
        if np.isnan(Rstar) : raise Exception('Stellar Radii is not added to \
            dataframe input or to planet_kwargs through the column/key named st_rad. Solar radii! \
            Please add it to one of them')

        #Mass in Jupiter Masses
        Mp = planet.get('pl_bmassj', planet_kwargs.get('pl_bmassj',np.nan))
        if np.isnan(Mp) : raise Exception('Planet Mass is not added to \
            dataframe input or to planet_kwargs through the column/key named pl_bmassj. J for JUPITER! \
            Please add it to one of them')  

        #TP profile params (3--Guillot 2010, Parmentier & Guillot 2013--see Line et al. 2013a for implementation)
        Tirr=planet.get('pl_eqt', planet_kwargs.get('pl_eqt',np.nan))

        if np.isnan(Tirr): 
            p =  planet.get('pl_orbper', planet_kwargs.get('pl_orbper',np.nan))
            p = p * (1*u.day).to(u.yr).value #convert to year 
            a =  (p**(2/3)*u.au).to(u.R_sun).value
            temp = planet.get('st_teff', planet_kwargs.get('st_teff',np.nan))
            Tirr = temp * np.sqrt(Rstar/(2*a))

        if np.isnan(Tirr): raise Exception('Planet Eq Temp is not added to \
            dataframe input or to planet_kwargs through the column/key named pl_eqt. Kelvin \
            Please add it to one of them') 

        p=planet.get('pl_orbper', planet_kwargs.get('pl_orbper',np.nan))

        if np.isnan(Tirr): raise Exception('Orbital Period is not added to \
            dataframe input or to planet_kwargs through the column/key named pl_orbper. Days Units') 
        else: 
            p = p * (1*u.day).to(u.yr).value #convert to year 
            a =  p**(2/3) #semi major axis in AU


        #setup picaso
        start_case = inputs()
        start_case.phase_angle(phase_angle) #radians 

        #define gravity
        start_case.gravity(mass=Mp, mass_unit=u.Unit('M_jup'),
                            radius=Rp, radius_unit=u.Unit('R_jup')) #any astropy units available

        #define star
        start_case.star(opacity, temp,logmh,logg,radius=Rstar, radius_unit=u.Unit('R_sun'),
                            semi_major=a, semi_major_unit=u.Unit('au'),
                            database = stellar_db ) #opacity db, pysynphot database, temp, metallicity, logg

        ##running this with all default inputs (users can override whatever after this initial run)
        start_case.guillot_pt(Tirr) 

    return start_case

def mean_regrid(x, y, newx=None, R=None):
    """
    Rebin the spectrum at a minimum R or on a fixed grid 

    Parameters
    ----------
    x : array 
        Wavenumbers
    y : array 
        Anything (e.g. albedo, flux)
    newx : array 
        new array to regrid on. 
    R : float 
        create grid with constant R

    Returns
    -------
    final x, and final y
    """
    if (isinstance(newx, type(None)) & (not isinstance(R, type(None)))) :
        newx = create_grid(1e4/max(x), 1e4/min(x), R)
    elif (not isinstance(newx, type(None)) & (isinstance(R, type(None)))) :  
        d = np.diff(newx)
        binedges = np.array([newx[0]-d[0]/2] + list(newx[0:-1]+d/2.0) + [newx[-1]+d[-1]/2])
        newx = binedges
    else: 
        raise Exception('Please either enter a newx or a R') 

    y, edges, binnum = binned_statistic(x,y,bins=newx)
    newx = (edges[0:-1]+edges[1:])/2.0

    return newx, y

def jupiter_pt():
    """Function to get Jupiter's PT profile"""
    return os.path.join(__refdata__, 'base_cases','jupiter.pt')
def jupiter_cld():
    """Function to get rough Jupiter Cloud model with fsed=3"""
    return os.path.join(__refdata__, 'base_cases','jupiterf3.cld')
def HJ_pt():
    """Function to get Jupiter's PT profile"""
    return os.path.join(__refdata__, 'base_cases','HJ.pt')
def HJ_pt_3d():
    """Function to get Jupiter's PT profile"""
    return os.path.join(__refdata__, 'base_cases','HJ_3d.pt')
def HJ_cld():
    """Function to get rough Jupiter Cloud model with fsed=3"""
    return os.path.join(__refdata__, 'base_cases','HJ.cld')
def single_phase_options(printout=True):
    """Retrieve all the options for direct radation"""
    if printout: print("Can also set functional form of forward/back scattering in approx['TTHG_params']")
    return ['cahoy','OTHG','TTHG','TTHG_ray']
def multi_phase_options(printout=True):
    """Retrieve all the options for multiple scattering radiation"""
    if printout: print("Can also set delta_eddington=True/False in approx['delta_eddington']")
    return ['N=2','N=1']
def raman_options():
    """Retrieve options for raman scattering approximtions"""
    return ["oklopcic","pollack","none"]



def all_planets():
    """
    Load all planets from https://exoplanetarchive.ipac.caltech.edu
    """
    # use this default URL to start out with 
    default_query = "https://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI?table=exoplanets&format=csv"
    all_planets =  requests.get(default_query)  
    all_planets = all_planets.text.replace(' ','').split('\n')

    # default data doesn't come with all the parameters we need so let's 
    # add those in 
    add_few_elements = '&select='+all_planets[0]+','.join(
        [',pl_trandur','pl_eqt','st_logg','st_metfe','st_j','st_h','st_k&'])

    # use requests to grab the csv formatted data and split by the commas 
    all_planets =  requests.get(default_query.split('&')[0] 
                                + add_few_elements + default_query.split('&')[1] )  
    all_planets = all_planets.text.replace(' ','').split('\n')

    # get into useful pandas dataframe format
    planets_df = pd.DataFrame(columns=all_planets[0].split(','), 
                             data = [i.split(',') for i in all_planets[1:-1]])
    planets_df = planets_df.replace(to_replace='', value=np.nan)    


    # convert to float when possible
    for i in planets_df.columns: 
        planets_df[i] = planets_df[i].astype(float,errors='ignore')

    return planets_df

def methodology_options(printout=True):
    """Retrieve all the options for methodology"""
    if printout: print("Can calculate spectrum using Toon 1989 methodology or sperhical harmonics")
    return ['Toon','SH']
def stream_options(printout=True):
    """Retrieve all the options for stream"""
    if printout: print("Can use 2-stream or 4-stream sperhical harmonics")
    return [2,4]




from .opacity_factory import ContinuumFactory ,MolecularFactory
import pandas as pd
import numpy as np
import os 
__refdata__ = os.environ.get('picaso_refdata')
def continuum(new_wno ='newwave_2000.coverage', original_file = 'CIA_DS_aug_2015.dat',
	colnames = ['wno','H2H2','H2He','H2H','H2CH4','H2N2'], new_filename=os.path.join(__refdata__, 'opacities','continuum.json')):
	"""
	This is a top level function to run the opacity facotry to compute the continuum. 
	It leverages D. Saumom's continuum opacity file and regrids it to the wavelength 
	that the user is attempting to create a spectrum on. 

	Parameters
	----------
	new_wno : str , array
		This is a pointer to the filename that is whitespace delimeted with at least one column 
		called "wavenumber". Can also be a numpy array
	original_file : str 
		This is a pointer to the original continuum opacity file that is structured according to 
		D. Saumon. This will eventually be replaced with the opacity db structure. 
	colnames : list, str
		This identifies the continuum absorbers that are located within the original_file with 
		opacities. In future updates, the file might be updated to include more continuum 
		absorbers. 
	"""
	if isinstance(new_wno ,str):
		new_wno = np.sort(pd.read_csv(new_wno, delim_whitespace=True)['wavenumber'].values)
	else: 
		new_wno = np.sort(new_wno)
	factory = ContinuumFactory(original_file,colnames, new_wno,new_filename=new_filename)
	factory.restructure_opacity()

def molecular(original_dir = '/Users/natashabatalha/Documents/AlbedoCodeWC/opacities/', 
	new_wno ='newwave_2000.coverage',new_filename=os.path.join(__refdata__, 'opacities','molecular.hdf5')):
	"""
	Currently, picaso is operating on a BYO opacities. Therefore, this function should be used 
	to go through a large directory of opacities and restructure them to the hdf5 databse that 
	picaso uses. 

	Warning
	-------
	This function is **not** yet general enough to be used by the public. It is highly tied to 
	the opacity grid calculated by R. Freedman and used by ExoARC group. There are several 
	subleties to how this database is created. For example, alkalis are added through a 
	separate script. Please contact 
	natasha.e.batalha@gmail.com if you have questions about opacities. 

	Parameters
	----------
	new_wno : str, array
		This is a pointer to the filename that is whitespace delimeted with at least one column 
		called "wavenumber". Can also be a numpy array.
	original_dier : str 
		This is a pointer to the a file that should have a directory for each molecule. 

	Todo
	----
	-Right now users have to add Na K CH4 from function `addNa&K&CH4.py` in scripts. Come up with a better method than when we 
	have a better opacity database
	"""
	
	if isinstance(new_wno ,str):
		new_wno = np.sort(pd.read_csv(new_wno, delim_whitespace=True)['wavenumber'].values)
	else: 
		new_wno = np.sort(new_wno)
	factory = MolecularFactory(original_dir, new_wno,new_filename=new_filename)
	factory.restructure_opacity()

#


import json
from .disco import get_angles_3d, compute_disco
import pandas as pd
import numpy as np
from scipy.spatial import cKDTree
import astropy.units as u
from .justdoit import inputs 
import pickle as pk
def rebin_mitgcm_pt(ng, nt, phase_angle, input_file, output_file,p_unit='Pa', kzz_unit = 'm*m/s',run_chem=False, MH=None, CtoO=None):
	"""
	Rebin GCM grid to a smaller grid. This function creates a pickle file 
	that can be directly input into `atmosphere_3d`.

	Parameters 
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle (radians)
	input_file : str
		PT input file you want to create the 3d grid from. Currently takes in a single 1d PT profile.
		Must have labeled, whitespace delimeted columns. In the columns, must have `temperature`, 
		`pressure`, and at least some molecular species (which are ALL case-sensitive)
	output_file : str
		output file location
	p_unit : str,optional
		Pressure Unit (default Pascal)
	kzz_unit : str,optional 
		Kzz Unit (default is m*m/s)
	run_chem : bool , optional
		(Optional) runs chemistry and adds to json file
	run_chem : bool , optional
		(Optional) MIT gcm output doesn't usually come with chemistry in the files 
		If this is the case then you can post process chemistry by setting this to True. 
	MH : float , optional
		(Optional) This is only used if run_chem is set to True. It is the Metallicity of 
		the planet in NON log units. MH = 1 is solar. 
	CtoO : float , optional
		(Optional) This is the C/O ratio of the planet also in NON log units. 
		CtoO = 0.55 is solar.  
	"""	

	gangle,gweight,tangle,tweight = get_angles_3d(ng,nt)
	ubar0, ubar1, cos_theta ,lat_picaso,lon_picaso = compute_disco(ng,nt, gangle, tangle, phase_angle)    


	threed_grid = pd.read_csv(input_file,delim_whitespace=True,names=['p','t','k'])
	all_lon= threed_grid.loc[np.isnan(threed_grid['k'])]['p'].values
	all_lat=  threed_grid.loc[np.isnan(threed_grid['k'])]['t'].values
	latlong_ind = np.concatenate((np.array(threed_grid.loc[np.isnan(threed_grid['k'])].index),[threed_grid.shape[0]] ))
	threed_grid = threed_grid.dropna() 

	lon = np.unique(all_lon)
	lat = np.unique(all_lat)

	nlon = len(lon)
	nlat = len(lat)
	total_pts = nlon*nlat
	nz = latlong_ind[1] - 1 

	p = np.zeros((nlon,nlat,nz))
	t = np.zeros((nlon,nlat,nz))
	kzz = np.zeros((nlon,nlat,nz))

	for i in range(len(latlong_ind)-1):

		ilon = list(lon).index(all_lon[i])
		ilat = list(lat).index(all_lat[i])

		p[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['p'].values
		t[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['t'].values
		kzz[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['k'].values


	lon2d, lat2d = np.meshgrid(lon_picaso, lat_picaso)
	lon2d = lon2d.flatten()
	lat2d = lat2d.flatten()

	xs, ys, zs = lon_lat_to_cartesian(all_lon,all_lat)
	xt, yt, zt = lon_lat_to_cartesian(lon2d,lat2d)

	tree = cKDTree(list(zip(xs,ys,zs)))
	nn = int(total_pts / (ng*nt))
	d,inds = tree.query(list(zip(xt,yt,zt)),k=nn)

	new_t = np.zeros((ng*nt,nz))
	new_kzz = np.zeros((ng*nt,nz))
	new_p = np.zeros((ng*nt,nz))

	for iz in range(0,nz):
		new_p[:,iz] = np.sum(p[:,:,iz].flatten()[inds],axis=1)/nn
		new_t[:,iz] = np.sum(t[:,:,iz].flatten()[inds],axis=1)/nn
		new_kzz[:,iz] = np.sum(kzz[:,:,iz].flatten()[inds],axis=1)/nn


	df = {} 
	df={int(i*180/np.pi):{} for i in lat_picaso}


	outfile = open(output_file,'w')
	if run_chem == True: case = inputs(chemeq=True)
	
	for ipos in range(0,ng*nt):
		
		#make sure units are always bars
		new_p = (new_p*u.Unit(p_unit)).to(u.bar).value
		new_kzz = (new_kzz*u.Unit(kzz_unit)).to('cm*cm/s').value

		if run_chem:
			case.inputs['atmosphere']['profile'] = pd.DataFrame({'temperature':new_t[ipos,0:nz], 'pressure':new_p[ipos,0:nz], 'kzz':new_kzz[ipos,0:nz]})
			case.chemeq(CtoO, MH)
			data = case.inputs['atmosphere']['profile']
		else : 
			data = pd.DataFrame({'temperature':new_t[ipos,0:nz], 'pressure':new_p[ipos,0:nz], 'kzz':new_kzz[ipos,0:nz]})

		df[int(lat2d[ipos]*180/np.pi)][int(lon2d[ipos]*180/np.pi)] = data

	df['phase_angle'] = phase_angle

	pk.dump(df,open(output_file, 'wb'))


def rebin_mitgcm_cld(ng, nt, phase_angle, input_file, output_file,names=['i','j','opd','g0','w0']):
	"""
	Rebin post processed GCM cloud grid to a smaller grid. This function creates a pickle file 
	that can be directly input into `clouds_3d`.

	Parameters 
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle (radians)
	input_file : str
		PT input file you want to create the 3d grid from. Currently takes in a single 1d PT profile.
		Must have labeled, whitespace delimeted columns. In the columns, must have `temperature`, 
		`pressure`, and at least some molecular species (which are ALL case-sensitive)
	output_file : str
		output file location
	names : list,str,optional
		List of names for the post processed file. The Default value is what generallly 
		comes out of A&M code.. But double check before running!! 
	"""	

	gangle,gweight,tangle,tweight = get_angles_3d(ng,nt)
	ubar0, ubar1, cos_theta ,lat_picaso,lon_picaso = compute_disco(ng,nt, gangle, tangle, phase_angle)    


	threed_grid = pd.read_csv(input_file,delim_whitespace=True,names=names)
	#get the lat and lon points by looking at the locations that the last 
	#index is nan but the first two have values. 
	#this assumes that the MIT GCM person has created their file by printing
	#Lon, Lat before the dump of output.cld
	all_lon= threed_grid.loc[np.isnan(threed_grid[names[-1]])][names[0]].values
	all_lat=  threed_grid.loc[np.isnan(threed_grid[names[-1]])][names[1]].values
	latlong_ind = np.concatenate((np.array(threed_grid.loc[np.isnan(threed_grid['k'])].index),[threed_grid.shape[0]] ))
	threed_grid = threed_grid.dropna() 

	lon = np.unique(all_lon)
	lat = np.unique(all_lat)

	nlon = len(lon)
	nlat = len(lat)
	total_pts = nlon*nlat
	nznw = latlong_ind[1] - 1 #number of wavepoints * number of levels

	opd = np.zeros((nlon,nlat,nznw))
	g0 = np.zeros((nlon,nlat,nznw))
	w0 = np.zeros((nlon,nlat,nznw))

	for i in range(len(latlong_ind)-1):

		ilon = list(lon).index(all_lon[i])
		ilat = list(lat).index(all_lat[i])

		opd[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['opd'].values
		g0[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['g0'].values
		w0[ilon, ilat, :] = threed_grid.loc[latlong_ind[i]:latlong_ind[i+1]]['w0'].values


	lon2d, lat2d = np.meshgrid(lon_picaso, lat_picaso)
	lon2d = lon2d.flatten()
	lat2d = lat2d.flatten()

	xs, ys, zs = lon_lat_to_cartesian(all_lon,all_lat)
	xt, yt, zt = lon_lat_to_cartesian(lon2d,lat2d)

	tree = cKDTree(list(zip(xs,ys,zs)))
	nn = int(total_pts / (ng*nt))
	d,inds = tree.query(list(zip(xt,yt,zt)),k=nn)

	new_opd = np.zeros((ng*nt,nznw))
	new_g0 = np.zeros((ng*nt,nznw))
	new_w0 = np.zeros((ng*nt,nznw))

	for iz in range(0,nznw):
		new_opd[:,iz] = np.sum(opd[:,:,iz].flatten()[inds],axis=1)/nn
		new_g0[:,iz] = np.sum(g0[:,:,iz].flatten()[inds],axis=1)/nn
		new_w0[:,iz] = np.sum(w0[:,:,iz].flatten()[inds],axis=1)/nn


	df = {} 
	df={int(i*180/np.pi):{} for i in lat_picaso}


	outfile = open(output_file,'w')
	if run_chem == True: case = inputs(chemeq=True)
	
	for ipos in range(0,ng*nt):
		
		data = pd.DataFrame({'opd':new_opd[ipos,0:nznw], 'g0':new_g0[ipos,0:nznw], 'w0':new_w0[ipos,0:nznw]})

		df[int(lat2d[ipos]*180/np.pi)][int(lon2d[ipos]*180/np.pi)] = data

	df['phase_angle'] = phase_angle

	pk.dump(df,open(output_file, 'wb'))

def make_3d_pt_input(ng,nt,phase_angle,input_file,output_file,**kwargs):
	"""Discontinued
	Program to create 3d PT input. Used to feed GCM input into disco ball.

	Parameters
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle
	input_file : str
		PT input file you want to create the 3d grid from. Currently takes in a single 1d PT profile.
		Must have labeled, whitespace delimeted columns. In the columns, must have `temperature`, 
		`pressure`, and at least some molecular species (which are ALL case-sensitive)
	output_file : str
		Output file location

	Returns
	-------
	Creates output file. No other returns.

	Examples
	--------
	Basic use of creating a 3D file: 

	>>> phase_angle = 0
	>>> number_gauss_angles = 2
	>>> number_tcheby_angles = 3
	>>> phase_angle = 0
	>>> input_file=jdi.jupiter_pt()
	>>> output_file='jupiter3D.hdf5'
	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file)
	
	Can also use `**kwargs` to control how pandas reads in input file 

	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file,usecols=[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,20,21])

	"""

	h5db = h5py.File(output_file,'w')
	#get geometry
	gangle,gweight,tangle,tweight = disco.get_angles_3d(ng, nt)
	ubar0, ubar1, cos_theta,lat,lon = disco.compute_disco(ng, nt, gangle, tangle, phase_angle)



	first = True
	dset = {}
	for g in gangle:
		for t in tangle:
			data = pd.read_csv(input_file, delim_whitespace=True,**kwargs)
			if first:
				dset['header'] = ','.join(list(data.keys()))
				first = False
			dset[str(g)][str(t)] = data

def make_3d_cld_input(ng,nt,phase_angle,input_file,output_file, lat_range=None, lon_range=None,rand_coverage=1):
	"""Discontinued
	Program to create 3d CLOUD input. Used to feed GCM input into disco ball.

	Parameters
	----------
	ng : int
		Number of gauss angles
	nt : int
		Number of Tchebysehv angles
	phase_angle : float
		Geometry of phase angle
	input_file : str
		CLD input file you want to create the 3d grid from. Currently takes in a single 1d CLD profile.
		Input file must be structured with the following columns in the correct order: ['lvl', 'wv','opd','g0','w0','sigma']
		input file 
	output_file : str
		Output file location
	lat_range : list
		(Optional)Range of latitudes to exclude in the CLD file
	lon_range : list
		(Optional)Range of longitudes to exclud in the CLD file
	rand_coverage : float 
		(Optional)Fractional cloud coverage. rand_cov=1 introduces full cloud coverage. rand_cov=0 has no
		coverage. 
	**kwargs : dict 
		Key word arguments used for `panads.read_csv`

	Returns
	-------
	Creates output file. No other returns.

	Examples
	--------
	Basic use of creating a 3D file: 
		
	>>> phase_angle = 0
	>>> number_gauss_angles = 2
	>>> number_tcheby_angles = 3
	>>> phase_angle = 0
	>>> input_file=jdi.jupiter_cld()
	>>> output_file='jupiter3D.hdf5'
	>>>make_3d_pt_input(number_gauss_angles, number_tcheby_angles,
	>>>         phase_angle, input_file,output_file)
	"""

	h5db = h5py.File(output_file,'w')
	#get geometry
	gangle,gweight,tangle,tweight = disco.get_angles_3d(ng, nt)
	ubar0, ubar1, cos_theta,lat,lon = disco.compute_disco(ng, nt, gangle, tangle, phase_angle)
	first = True

	for g, lg in zip(gangle,lon):
		for t, lt in zip(tangle, lat):
			data = pd.read_csv(input_file, delim_whitespace = True,
					header=None, skiprows=1, names = ['lvl', 'wv','opd','g0','w0','sigma'],
					dtype='f8')
			if not isinstance(lat_range,type(None)):
				if (lt > np.min(lat_range)) and (lt<np.max(lat_range)):
					data = data*0
			if not isinstance(lon_range,type(None)):
				if (lg > np.min(lon_range)) and (lg < np.max(lon_range)):
					data = data*0
			if np.random.rand() > rand_coverage:
				data = data*0

			if first:
				h5db.attrs['header'] = ','.join(list(data.keys()))
				first = False
			dset = h5db.create_dataset(str(g)+'/'+str(t), data=data, chunks=True)

def lon_lat_to_cartesian(lon, lat, R = 1):
	"""
	Calculates lon, lat coordinates of a point on a sphere with
	radius R

	Parameters 
	----------
	lon : array
		Longitude
	lat : array 
		Latitude 
	R : float	
		radius of sphere 

	Returns
	-------
	x, y, z 
	"""
	x =  R * np.cos(lat) * np.cos(lon)
	y = R * np.cos(lat) * np.sin(lon)
	z = R * np.sin(lat)
	return x,y,z
